from collections.abc import Sequence
from typing import Optional

from qibolab._core.identifier import ChannelId
from qibolab._core.instruments.qblox.sequence.asm import label
from qibolab._core.pulses.pulse import PulseId
from qibolab._core.serialize import Model
from qibolab._core.sweeper import ParallelSweepers, iteration_length

from ..q1asm.ast_ import (
    Add,
    Block,
    BlockIter,
    BlockList,
    Instruction,
    Jlt,
    Line,
    Loop,
    Move,
    Nop,
    Reference,
    Register,
    ResetPh,
    Sub,
    Wait,
)
from .asm import Registers
from .sweepers import IndexedParams, Param, update_instructions

__all__ = []


class LoopSpec(Model):
    """Loop descriptor.

    Created by the :func:`loops` function.
    """

    reg: Register
    """Register used for the loop counter."""
    length: int
    """Iteration length."""
    id: Optional[int]
    """Iteration index.

    `None` marks the loop as the shots loop.
    """

    @property
    def shots(self) -> bool:
        return self.id is None

    @property
    def description(self) -> str:
        """Sweeper textual description."""
        return f"loop {self.id + 1}" if self.id is not None else "shots"


def loops(
    sweepers: list[ParallelSweepers], nshots: int, inner_shots: bool
) -> Sequence[LoopSpec]:
    """Initialize registers for loop counters.

    The counters implement the ``length`` of the iteration, which, for a general
    sweeper, is fully characterized by a ``(start, step, length)`` tuple.

    Those related to :attr:`Registers.bin` and :attr:`Registers.bin_reset` are actually
    not loop counter on their own, but they are required to properly store the
    acquisitions in different bins.
    """
    shots = LoopSpec(reg=Registers.shots.value, length=nshots, id=None)
    sweep = [
        LoopSpec(
            reg=Register(number=i + Registers.first_available()),
            length=iteration_length(parsweep),
            id=i,
        )
        # the first sweeper should be the outermost, thus reverse them during the
        # enumeration
        for i, parsweep in enumerate(sweepers[::-1])
    ]
    return [shots] + sweep if inner_shots else sweep + [shots]


START: str = "start"
"""Label first experiment's instruction."""
SHOTS: str = "shots"
"""Label instruction after bin counter reset instruction.

The instruction is used for shots, and skipped when all shots are
finished.

Cf. :const:`_SHOTS_BIN_RESET`.
"""


def _experiment_end(relaxation_time: int) -> list[Line]:
    """Wrap up experiment.

    - relax
    - reset phase (?)
    - increment bin where the result is saved

    The bin increment is possibly reset later on, in order to save on the same bin, and
    thus summing various shots (eventually averaging). Cf. :const:`_SHOTS_BIN_RESET`.
    """
    return [
        Line(instruction=Wait(duration=relaxation_time), comment="relaxation"),
        Line(instruction=ResetPh(), comment="phase reset"),
        Line(
            instruction=Add(
                a=Registers.bin.value, b=1, destination=Registers.bin.value
            ),
            comment="bin increment",
        ),
    ]


_SHOTS_BIN_RESET: list[Line] = [
    Line(
        instruction=Jlt(a=Registers.shots.value, b=1, address=Reference(label=SHOTS)),
        comment="skip bin reset - advance both counters",
    ),
    Line(
        instruction=Move(
            source=Registers.bin_reset.value,
            destination=Registers.bin.value,
        ),
        comment="shots average: reset bin counter",
    ),
]
"""Reset bin counter to save shots to the same one.

The full algorithm is detailed at:
https://github.com/qiboteam/qibolab/discussions/1119
"""


def _sweep_update(p: Param, channel: set[ChannelId], pulses: set[PulseId]) -> Block:
    """Sweeper update for a single parameter.

    - increment the parameter register
    - set new parameter value (if channel-wise)
        - an additional `nop` instruction is plugged to wait one further clock cycle in
          between the register increment and the parameter update, to ensure the value
          is correctly propagated
          https://docs.qblox.com/en/main/products/architecture/sequencers/sequencer.html#registers
    """
    return (
        *(
            (
                Line(
                    instruction=(
                        Add(a=p.reg, b=p.step, destination=p.reg)
                        if p.step >= 0
                        else Sub(a=p.reg, b=-p.step, destination=p.reg)
                    ),
                    comment=f"shift {p.description}",
                ),
            )
            if p.channel in channel or p.pulse in pulses
            else ()
        ),
        *(
            (
                # wait one more clock cycle
                [Nop()]
                # then update the value
                + update_instructions(p.role, p.reg)
            )
            if p.channel in channel
            else ()
        ),
    )


def _sweep_updates(
    lp: LoopSpec, params: IndexedParams, channel: set[ChannelId], pulses: set[PulseId]
) -> BlockIter:
    """Parallel sweeper updates.

    Collects all synced updates, those happening in a single loop.
    """
    return (
        (
            line
            for block in (
                _sweep_update(p, channel, pulses) for p in (params[lp.id].all)
            )
            for line in block
        )
        if lp.id is not None
        else ()
    )


def _sweep_reset(
    params: list[Param], channel: set[ChannelId], pulses: set[PulseId]
) -> BlockList:
    """Reset sweeper register value.

    Once the loop is completed, the parameter value, which is hold in the respective
    register, needs to be reset to its original value. To be ready when a new loop will
    possibly start (which is always the case, but for the outermost loop).

    .. note::

        Channel parameters are also updated immediately, in order for the change to take
        effect.
        Pulse parameters will anyhow act around the suitable pulse, so the update is
        always performed when needed.
    """
    return (
        [
            Line(
                instruction=Move(source=p.start, destination=p.reg),
                comment=f"init {p.description}",
            )
            for p in params
            if p.channel in channel or p.pulse in pulses
        ]
        # wait one clock cycle before parameters' update
        # cf. _sweep_update()
        + [Nop()]
        + [
            inst
            for p in params
            if p.channel in channel
            for inst in update_instructions(p.role, p.reg)
        ]
    )


def _sweep_iteration(
    lp: LoopSpec,
    params: IndexedParams,
    shots: bool,
    channel: set[ChannelId],
    pulses: set[PulseId],
) -> BlockList:
    """Sweep loop.

    The operations performed by the looping machinery are divided in internal to loop,
    and those happening right after completing all loops.

    The internal ones consist of:

    - update the parameters for all the sweepers set at the selected loop level
    - loop, i.e. decrement iteration counter and jump back to experiment start

    While the closing operations include:

    - reset iteration counter, after a whole cycle is completed
    - reset the parameters' values, to possibly get ready for an entire new iteration,
      triggered by an external loop
    """
    loops_ = [
        *_sweep_updates(lp, params, channel, pulses),
        Line(
            instruction=Loop(a=lp.reg, address=Reference(label=START)),
            comment=f"loop over {lp.description}",
            label=SHOTS if shots else None,
        ),
        Move(source=lp.length, destination=lp.reg),
    ]
    # no parameter reset for the shots loop, since, by definition, that's just an unaltered
    # repetition of the experiment
    reset = [] if lp.id is None else _sweep_reset(params[lp.id].all, channel, pulses)
    return loops_ + reset


def _loop_machinery(
    loops: Sequence[LoopSpec],
    params: IndexedParams,
    singleshot: bool,
    channel: set[ChannelId],
    pulses: set[PulseId],
) -> BlockList:
    """Looping block.

    It creates an instruction block to be entirely placed after the
    experiment (including the relaxation part) to repeat it as
    specified, taking into account both shots and nested sweeper loops.
    """

    def shots(marker: LoopSpec) -> bool:
        return marker.shots and not singleshot

    return [
        inst
        for lp in loops
        for inst in (
            (_SHOTS_BIN_RESET if shots(lp) else [])
            + _sweep_iteration(lp, params, shots(lp), channel, pulses)
        )
    ][:-1]


def loop(
    experiment: list[Instruction],
    loops: Sequence[LoopSpec],
    params: IndexedParams,
    relaxation_time: int,
    singleshot: bool,
    channel: set[ChannelId],
    pulses: set[PulseId],
) -> Block:
    end = _experiment_end(relaxation_time)
    machinery = _loop_machinery(loops, params, singleshot, channel, pulses)
    main = experiment + end + machinery

    return [label(main[0], START)] + main[1:]



"""Emulator controller."""

from collections import defaultdict
from collections.abc import Iterable
from functools import reduce
from operator import or_
from typing import Callable, Optional, cast

import numpy as np
from numpy.typing import NDArray

from qibolab._core.components import Config
from qibolab._core.components.configs import AcquisitionConfig
from qibolab._core.execution_parameters import ExecutionParameters
from qibolab._core.identifier import Result
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import (
    Acquisition,
    Delay,
    Pulse,
    PulseLike,
    Readout,
    VirtualZ,
)
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers

from .engine import Operator, OperatorEvolution, QutipEngine, SimulationEngine
from .hamiltonians import (
    HamiltonianConfig,
    Modulated,
    waveform,
)
from .results import acquisitions, index, results, select_acquisitions

__all__ = ["EmulatorController"]


class EmulatorController(Controller):
    """Emulator controller."""

    engine: SimulationEngine = QutipEngine()
    """SimulationEngine. Default is QutipEngine."""
    bounds: str = "emulator/bounds"

    def connect(self):
        """Dummy connect method."""

    def disconnect(self):
        """Dummy disconnect method."""

    @property
    def sampling_rate(self):
        """Sampling rate of emulator."""
        return 1

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ) -> dict[int, Result]:
        # convert align to delays
        sequences_ = (seq.align_to_delays() for seq in sequences)
        # just merge the results of multiple executions in a single dictionary
        return reduce(
            or_,
            (
                results(
                    # states in computational basis
                    self._sweep(sequence, configs, sweepers),
                    sequence,
                    cast(HamiltonianConfig, configs["hamiltonian"]),
                    options,
                )
                for sequence in sequences_
            ),
        )

    def _sweep(
        self,
        sequence: PulseSequence,
        configs: dict[str, Config],
        sweepers: list[ParallelSweepers],
        updates: Optional[dict] = None,
    ) -> NDArray:
        """Sweep over sequence.

        This function invokes itself recursively, adding an array
        dimension at each call as the outermost one. The extra dimension
        corresponds to the values in the first nested sweep (with the
        lowest index, interpreted as the outermost as well).
        """
        # use a default dictionary, merging existing values
        updates = defaultdict(dict) | ({} if updates is None else updates)

        if len(sweepers) == 0:
            return self._play_sequence(sequence, configs, updates)

        parsweep = sweepers[0]
        # collect slices of results, corresponding to the current iteration
        results = []
        # execute once for each parallel value
        for values in zip(*(s.values for s in parsweep)):
            # update all parallel sweepers with the respective values
            for sweeper, value in zip(parsweep, values):
                if sweeper.pulses is not None:
                    for pulse in sweeper.pulses:
                        updates[pulse.id].update({sweeper.parameter.name: value})
                if sweeper.channels is not None:
                    for channel in sweeper.channels:
                        updates[channel].update({sweeper.parameter.name: value})

            # append new slice for the current parallel value
            results.append(self._sweep(sequence, configs, sweepers[1:], updates))

        # stack all slices in a single array, along the current outermost dimension
        return np.stack(results)

    def _play_sequence(
        self, sequence: PulseSequence, configs: dict[str, Config], updates: dict
    ) -> NDArray:
        """Play single sequence on emulator.

        The array returned by this function has a single dimension, over
        the various measurements included in the sequence.
        """
        sequence_ = update_sequence(sequence, updates)
        tlist_ = tlist(sequence_, self.sampling_rate)
        configs_ = update_configs(configs, updates)
        config = cast(HamiltonianConfig, configs_["hamiltonian"])
        hamiltonian = config.hamiltonian(config=configs_, engine=self.engine)
        time_hamiltonian = self._pulse_hamiltonian(sequence_, configs_)
        results = self.engine.evolve(
            hamiltonian=hamiltonian,
            initial_state=config.initial_state(self.engine),
            time=tlist_,
            collapse_operators=config.dissipation(self.engine),
            time_hamiltonian=time_hamiltonian,
        )
        return select_acquisitions(
            results.states,
            acquisitions(sequence_).values(),
            tlist_,
        )

    def _pulse_hamiltonian(
        self, sequence: PulseSequence, configs: dict[str, Config]
    ) -> Optional[OperatorEvolution]:
        """Construct Hamiltonian time dependent term for qutip simulation."""

        channels = [
            [operator, channel_time(waveforms)]
            for operator, waveforms in hamiltonians(sequence, configs, self.engine)
        ]
        return OperatorEvolution(channels) if len(channels) > 0 else None


def update_sequence(sequence: PulseSequence, updates: dict) -> PulseSequence:
    """Apply sweep updates to base sequence."""
    return PulseSequence(
        [(ch, e.model_copy(update=updates.get(e.id, {}))) for ch, e in sequence]
    )


def update_configs(configs: dict[str, Config], updates: dict) -> dict[str, Config]:
    """Apply sweep updates to base configs."""
    return {k: c.model_copy(update=updates.get(k, {})) for k, c in configs.items()}


def tlist(
    sequence: PulseSequence, sampling_rate: float, per_sample: float = 2
) -> NDArray:
    """Compute times for evolution.

    The frequency of times is double the sampling rate, to make sure
    that all pulses features are resolved by the evolution.

    This can be customized using the `per_sample` rate, e.g. to retrieve times at the
    sampling rate itself, for pulses evaluation.

    .. note::

        As an optimization, if an acquisition is executed as the last
        sequence operation, that's not taken into account, since it is not
        simulated by the present emulator.

        For long experiments, it is a mild optimization. But it critically speeds up
        short experiments, given the usual relative duration of acquisitions and control
        pulses.
    """
    seq = (
        sequence[:-1]
        if isinstance(sequence[-1][1], (Acquisition, Readout))
        else sequence
    )
    end = max(seq.duration, 1)
    rate = sampling_rate * per_sample
    return np.arange(0, end, 1 / rate)


def hamiltonian(
    pulses: Iterable[PulseLike],
    config: Config,
    hamiltonian: HamiltonianConfig,
    hilbert_space_index: int,
    engine: SimulationEngine,
) -> tuple[Operator, list[Modulated]]:
    n = hamiltonian.transmon_levels
    op = engine.expand(
        config.operator(n=n, engine=engine), hamiltonian.dims, hilbert_space_index
    )
    waveforms = (
        waveform(pulse, config, hamiltonian.qubits[hilbert_space_index])
        for pulse in pulses
        if isinstance(pulse, (Pulse, Delay, VirtualZ))
    )
    return (op, [w for w in waveforms if w is not None])


def hamiltonians(
    sequence: PulseSequence, configs: dict[str, Config], engine: SimulationEngine
) -> Iterable[tuple[Operator, list[Modulated]]]:
    hconfig = cast(HamiltonianConfig, configs["hamiltonian"])
    return (
        hamiltonian(
            sequence.channel(ch),
            configs[ch],
            hconfig,
            index(ch, hconfig),
            engine,
        )
        for ch in sequence.channels
        # TODO: drop the following, and treat acquisitions just as empty channels
        if not isinstance(configs[ch], AcquisitionConfig)
    )


def channel_time(waveforms: Iterable[Modulated]) -> Callable[[float], float]:
    """Wrap time function for specific channel.

    Used to avoid late binding issues.
    """

    def time(t: float) -> float:
        cumulative_time = 0
        cumulative_phase = 0
        for pulse in waveforms:
            pulse_duration = pulse.duration  # TODO: pass sampling rate
            pulse_phase = pulse.phase
            if cumulative_time <= t < cumulative_time + pulse_duration:
                relative_time = t - cumulative_time
                index = int(relative_time)  # TODO: pass sampling rate
                return pulse(t, index, cumulative_phase)
            cumulative_time += pulse_duration
            cumulative_phase += pulse_phase
        return 0

    return time



from typing import Optional

from qibolab._core.pulses.pulse import (
    Acquisition,
    Align,
    Delay,
    Pulse,
    Readout,
    VirtualZ,
)
from qibolab._core.sweeper import Parameter

from ..q1asm.ast_ import (
    Acquire,
    Block,
    Instruction,
    Line,
    Play,
    Register,
    SetPhDelta,
    UpdParam,
    Wait,
    WaitSync,
)
from .acquisition import AcquisitionSpec, MeasureId
from .asm import Registers, convert
from .sweepers import (
    ParameterizedPulse,
    ParamRole,
    SweepSequence,
    reset_instructions,
    update_instructions,
)
from .waveforms import WaveformIndices

__all__ = []


def play_pulse(pulse: Pulse, waveforms: WaveformIndices) -> Line:
    uid = pulse.id
    w0 = waveforms[(uid, 0)]
    w1 = waveforms[(uid, 1)]
    assert w0[1] == w1[1]
    return Line(
        instruction=Play(wave_0=w0[0], wave_1=w1[0], duration=w0[1]),
        comment=f"id: 0x{uid.hex[:5]}",
    )


def play_duration_swept(registers: dict[ParamRole, Register]) -> list[Instruction]:
    return [
        Play(
            wave_0=registers[ParamRole.PULSE_I],
            wave_1=registers[ParamRole.PULSE_Q],
            duration=4,
        ),
        Wait(duration=registers[ParamRole.DURATION]),
    ]


def play(
    parpulse: ParameterizedPulse,
    waveforms: WaveformIndices,
    acquisitions: dict[MeasureId, AcquisitionSpec],
    time_of_flight: Optional[float],
) -> Block:
    """Process the individual pulse in experiment."""
    pulse = parpulse[0]
    params = parpulse[1]

    if isinstance(pulse, Pulse):
        phase = int(convert(pulse.relative_phase, Parameter.relative_phase))
        minus_phase = int(convert(-pulse.relative_phase, Parameter.relative_phase))
        duration_sweep = {
            p.role: p.reg for p in params if p.role.value[1] is Parameter.duration
        }
        return (
            ([SetPhDelta(value=phase)] if phase != 0 else [])
            + (
                [play_pulse(pulse, waveforms)]
                if len(duration_sweep) == 0
                else play_duration_swept(duration_sweep)
            )
            + (
                [SetPhDelta(value=minus_phase), UpdParam(duration=4)]
                if phase != 0
                else []
            )
        )
    if isinstance(pulse, Delay):
        return [
            Line(
                instruction=Wait(duration=int(pulse.duration)),
                comment=f"id: 0x{pulse.id.hex[:5]}",
            )
            if len(params) == 0
            else Wait(duration=next(iter(params)).reg)
        ]
    if isinstance(pulse, VirtualZ):
        return [
            SetPhDelta(
                value=int(convert(pulse.phase, Parameter.relative_phase))
                if len(params) == 0
                else next(iter(params)).reg
            ),
            UpdParam(duration=4),
        ]
    if isinstance(pulse, Acquisition):
        acq = acquisitions[pulse.id]
        return [
            Acquire(
                acquisition=acq.acquisition.index,
                bin=Registers.bin.value,
                duration=acq.duration,
            )
        ]
    if isinstance(pulse, Align):
        raise NotImplementedError("Align operation not yet supported by Qblox.")
    if isinstance(pulse, Readout):
        acq = acquisitions[pulse.id]
        delay = int(time_of_flight) if time_of_flight is not None else 4
        return [
            play_pulse(pulse.probe, waveforms).update({"duration": delay}),
            Acquire(
                acquisition=acq.acquisition.index,
                bin=Registers.bin.value,
                duration=int(pulse.duration),
            ),
        ]
    raise NotImplementedError(f"Instruction {type(pulse)} unsupported by Qblox driver.")


def event(
    parpulse: ParameterizedPulse,
    waveforms: WaveformIndices,
    acquisitions: dict[MeasureId, AcquisitionSpec],
    time_of_flight: Optional[float],
) -> Block:
    params = parpulse[1]
    return [
        inst
        for block in (
            *(update_instructions(p.role, p.reg) for p in params),
            *(play(parpulse, waveforms, acquisitions, time_of_flight),),
            *(reset_instructions(p.role, p.reg) for p in reversed(list(params))),
        )
        for inst in block
    ]


def experiment(
    sequence: SweepSequence,
    waveforms: WaveformIndices,
    acquisitions: dict[MeasureId, AcquisitionSpec],
    time_of_flight: Optional[float],
) -> Block:
    """Representation of the actual experiment to be executed.

    The parameters' update (`upd_param`) in front of everything is needed to ensure that
    the parameter values for sweepers targeting channels have been updated. The updates
    for those targeting pulses will be triggered by the `play` instruction (and they
    *have to* be local).

    The synchronization (`wait_sync`) will guarantee the common start of all involved
    channels, which is otherwise hard to control (and debug).
    """
    return [UpdParam(duration=4), WaitSync(duration=4)] + [
        inst
        for block in (
            event(pulse, waveforms, acquisitions, time_of_flight) for pulse in sequence
        )
        for inst in block
    ]



from typing import List

import numpy as np
import numpy.typing as npt
from scipy.linalg import toeplitz

from .abstract import Extractor

__all__ = ["ToeplitzExtractor"]


def unpackbits(x: npt.NDArray, num_bits: int) -> npt.NDArray:
    if np.issubdtype(x.dtype, np.floating):
        raise ValueError("numpy data type needs to be int-like")
    xshape = list(x.shape)
    x = x.reshape([-1, 1])
    mask = 2 ** np.arange(num_bits, dtype=x.dtype).reshape([1, num_bits])
    return (x & mask).astype(bool).astype(int).reshape(xshape + [num_bits])


def generate_toeplitz(input_bits: int, extraction_ratio: int) -> npt.NDArray:
    """Generate a pseudo-random Toeplitz matrix of dimension ``(input_bits, extraction_ratio)``."""
    while True:
        c = np.mod(np.random.permutation(input_bits), 2)
        r = np.mod(np.random.permutation(extraction_ratio), 2)
        if np.sum(c) == 6 and np.sum(r) == (extraction_ratio // 2):
            return toeplitz(c, r)


def toeplitz_extract(
    m1: npt.NDArray, input_bits: int, extraction_ratio: int
) -> npt.NDArray:
    m2 = unpackbits(m1, input_bits)
    m2 = np.flip(m2)

    m3 = generate_toeplitz(input_bits, extraction_ratio)

    m4 = np.matmul(m2, m3)
    m4 = m4.astype("int16")
    m4 = np.mod(m4, 2)
    m4 = np.packbits(m4)
    m4 = np.right_shift(m4, 8 - extraction_ratio)
    return m4


def upscale(samples: npt.NDArray, input_bits=4, output_bits=32) -> npt.NDArray:
    """Increase size of random bit strings by concatenating them."""
    assert output_bits > input_bits
    assert output_bits % input_bits == 0

    factor = output_bits // input_bits
    assert len(samples) % factor == 0

    n = len(samples) // factor
    upscaled = np.zeros(n, dtype=int)
    for i, s in enumerate(np.reshape(samples, (factor, n))):
        upscaled += s << (input_bits * i)
    return upscaled


class ToeplitzExtractor(Extractor):
    """https://arxiv.org/pdf/2402.09481 appendix A.5"""

    input_bits: int = 12
    """Number of bits of the raw numbers sampled from the QRNG."""
    extraction_ratio: int = 4
    """Number of bits of the uniformly distributed extracted output samples."""
    precision_bits: int = 32

    def __post_init__(self):
        if self.precision_bits % self.extraction_ratio != 0:
            raise ValueError(
                f"Number of bits must be a multiple of the extracted bits {self.extraction_ratio}."
            )

    def num_raw_samples(self, n: int) -> int:
        return 2 * n * (self.precision_bits // self.extraction_ratio)

    def extract(self, raw: List[int]) -> npt.NDArray:
        extracted = toeplitz_extract(
            np.array(raw), self.input_bits, self.extraction_ratio
        ).astype(int)
        upscaled = upscale(extracted, self.extraction_ratio, self.precision_bits)
        return upscaled.astype(float) / (2**self.precision_bits - 1)



from pathlib import Path
from typing import Any, Optional

from qblox_instruments import Cluster

from qibolab._core.components.configs import Configs, LogConfig
from qibolab._core.identifier import ChannelId

from .identifiers import SequencerMap
from .results import AcquiredData
from .sequence import Q1Sequence

__all__ = []


def _check(configs: Configs) -> Optional[Path]:
    if "log" in configs:
        assert isinstance(configs["log"], LogConfig)
        return configs["log"].path


def _sanitize(name: str) -> str:
    return name.replace("/", "-")


class Logger:
    def __init__(self, configs: Configs) -> None:
        self.path = _check(configs)

    def __getattribute__(self, name: str, /) -> Any:
        if super().__getattribute__("path") is None:
            return lambda *args, **kwargs: None
        return super().__getattribute__(name)

    def sequences(self, sequences: dict[ChannelId, Q1Sequence]):
        assert self.path is not None
        for ch, seq in sequences.items():
            (self.path / _sanitize(f"{ch}.json")).write_text(seq.model_dump_json())

    def status(self, cluster: Cluster, sequencers: SequencerMap):
        assert self.path is not None
        status = self.path / "status"
        status.mkdir(exist_ok=True)

        (status / "cluster.json").write_text(str(cluster.snapshot()))
        for slot, seqs in sequencers.items():
            for ch, seq_idx in seqs.items():
                (status / _sanitize(f"{ch}.log")).write_text(
                    str(cluster.get_sequencer_status(slot, seq_idx))
                )

    def data(self, data: dict[ChannelId, AcquiredData]):
        pass



"""Convert helper functions for qibosoq driver."""

from dataclasses import asdict
from functools import singledispatch
from typing import Any

import numpy as np
import qibosoq.components.base as rfsoc
import qibosoq.components.pulses as rfsoc_pulses
from scipy.constants import mega, micro, nano

from qibolab._core.components.channels import (
    AcquisitionChannel,
    Channel,
    DcChannel,
    IqChannel,
)
from qibolab._core.components.configs import Config, OscillatorConfig
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses.envelope import (
    Drag,
    Envelope,
    Exponential,
    Gaussian,
    Rectangular,
)
from qibolab._core.pulses.pulse import (
    Acquisition,
    Align,
    Delay,
    PulseId,
    PulseLike,
    Readout,
    VirtualZ,
)
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import Parameter, Sweeper


def replace_pulse_shape(
    rfsoc_pulse: rfsoc_pulses.Pulse, envelope: Envelope, samples: int
) -> rfsoc_pulses.Pulse:
    """Set pulse shape parameters in rfsoc_pulses pulse object."""
    if isinstance(envelope, Rectangular):
        return rfsoc_pulses.Rectangular(**asdict(rfsoc_pulse))
    if isinstance(envelope, Gaussian):
        return rfsoc_pulses.Gaussian(
            **asdict(rfsoc_pulse), rel_sigma=1 / envelope.rel_sigma
        )
    if isinstance(envelope, Drag):
        return rfsoc_pulses.Drag(
            **asdict(rfsoc_pulse), rel_sigma=1 / envelope.rel_sigma, beta=envelope.beta
        )
    if isinstance(envelope, Exponential):
        return rfsoc_pulses.FluxExponential(
            **asdict(rfsoc_pulse),
            tau=envelope.tau,
            upsilon=envelope.upsilon,
            weight=envelope.g,
        )

    return rfsoc_pulses.Arbitrary(
        **asdict(rfsoc_pulse),
        i_values=envelope.i(samples).tolist(),
        q_values=envelope.q(samples).tolist(),
    )


def get_lo_frequency(ch: Channel, configs: dict[str, Config]) -> float:
    """Return LO frequency from channel, if applicable."""
    if isinstance(ch, IqChannel):
        if ch.lo is not None:
            conf = configs[ch.lo]
            if isinstance(conf, OscillatorConfig):
                return conf.frequency
    return 0


def convert_units_sweeper(
    sweeper: Sweeper,
    channels: dict[ChannelId, Channel],
    configs: dict[str, Config],
) -> Sweeper:
    """Convert units for `qibosoq.abstract.Sweeper` considering also LOs."""

    start, stop, step = sweeper.irange
    new_start, new_stop, new_step = sweeper.irange
    if sweeper.parameter is Parameter.frequency:
        assert sweeper.channels is not None
        lo_frequency = get_lo_frequency(channels[sweeper.channels[0]], configs)

        new_start = (start - lo_frequency) / mega
        new_stop = (stop - lo_frequency) / mega
        new_step = step / mega

    elif sweeper.parameter in (Parameter.phase, Parameter.relative_phase):
        new_start = np.degrees(start)
        new_stop = np.degrees(stop)
        new_step = np.degrees(step)

    elif sweeper.parameter is Parameter.duration:
        new_start = start / micro * nano
        new_stop = stop / micro * nano
        new_step = step / micro * nano

    return Sweeper(
        parameter=sweeper.parameter,
        range=(new_start, new_stop, new_step),
        pulses=sweeper.pulses,
        channels=sweeper.channels,
    )


def order_pulse_sequence(
    sequence: PulseSequence,
) -> tuple[list[tuple[ChannelId, PulseLike, float]], dict[PulseId, list[PulseId]]]:
    """Order pulse sequence by execution time."""

    channel_time: dict[str, float] = {}
    channel_order = {ch: i for i, ch in enumerate(sequence.by_channel.keys())}
    result = []

    delay_equivalence = {}

    for ch, subseq in sequence.by_channel.items():
        channel_time[ch] = 0
        delays_before_pulse = []
        for idx, pulse in enumerate(subseq):
            if isinstance(pulse, Delay):
                start = channel_time[ch]
                channel_time[ch] = stop = start + pulse.duration
                delays_before_pulse.append(pulse.id)

            elif isinstance(pulse, Align):
                start = channel_time[ch]
                stop = max(channel_time.values())
                for c in channel_time:
                    channel_time[c] = stop
                delays_before_pulse.append(pulse.id)

            else:
                start = channel_time[ch]
                channel_time[ch] = stop = start + pulse.duration
                result.append((ch, pulse, start, stop, idx))
                if len(delays_before_pulse) > 0:
                    delay_equivalence[pulse.id] = delays_before_pulse
                delays_before_pulse = []

    def sort_key(x):
        """Sort pulse first by time and then by type (delays/align later)."""
        ch, _, start, stop, idx = x
        return (start, stop, idx, channel_order[ch])

    result.sort(key=sort_key)

    rt_result = []
    rst = 0  # relative start time
    last_start = 0
    for ch, pulse, start, stop, _ in result:
        rst = start - last_start
        rt_result.append((ch, pulse, rst))
        last_start = start

    return rt_result, delay_equivalence


def simplify_delays(
    timed_sequence: list[tuple[ChannelId, PulseLike, float]],
) -> tuple[list[tuple[ChannelId, PulseLike, float]], dict[PulseId, list[PulseId]]]:
    """Merge consecutive delays at the same start time by subtracting the minimum delay."""
    compressed = []
    delay_equivalence = {}
    last_id = None

    last_delay = 0
    for ch, pulse, t in timed_sequence:
        if not isinstance(pulse, Delay):
            compressed.append((ch, pulse, t))
            last_delay = 0
            last_id = None
        elif pulse.duration <= last_delay:
            if last_id is None:
                delay_equivalence[pulse.id] = [pulse.id]
                last_id = pulse.id
            else:
                delay_equivalence[last_id].append(pulse.id)
        elif pulse.duration > last_delay:
            new_duration = pulse.duration - last_delay
            last_delay = pulse.duration
            pulse = Delay(duration=new_duration, id_=pulse.id)
            compressed.append((ch, pulse, t))

            if last_id is None:
                delay_equivalence[pulse.id] = [pulse.id]
                last_id = pulse.id
            else:
                delay_equivalence[last_id].append(pulse.id)
    return compressed, delay_equivalence


def get_index(sequence: list[PulseLike], pulse: PulseLike) -> int:
    """Get pulse index from a sequence."""
    for idx, p in enumerate(sequence):
        if p.id == pulse.id:
            return idx
    raise RuntimeError("Pulse not in sequence.")


@singledispatch
def convert(*args) -> Any:
    """Convert from qibolab obj to qibosoq obj, overloaded."""
    raise ValueError(f"Convert function received bad parameters ({type(args[0])}).")


@convert.register
def _(
    sequence: PulseSequence,
    sampling_rate: float,
    channels: dict[ChannelId, Channel],
    configs: dict[str, Config],
) -> list[dict]:
    """Convert PulseSequence to list of rfosc pulses with relative time."""

    list_sequence = []
    ordered_sequence, _ = order_pulse_sequence(sequence)

    ch_vz_phases = {}

    for ch, pulse, rst in ordered_sequence:
        if ch not in ch_vz_phases:
            ch_vz_phases[ch] = 0.0
        if isinstance(pulse, VirtualZ):
            ch_vz_phases[ch] += pulse.phase
            continue

        pulse = convert(
            pulse,
            rst * nano / micro,
            ch,
            channels,
            sampling_rate,
            configs,
            ch_vz_phases[ch],
        )
        pulse_dict = asdict(pulse)
        list_sequence.append(pulse_dict)

    return list_sequence


@convert.register
def _(
    pulse: PulseLike,
    start_delay: float,
    ch_id: ChannelId,
    channels: dict[ChannelId, Channel],
    sampling_rate: float,
    configs: dict[str, Config],
    ch_vz_phase: float,
) -> rfsoc_pulses.Element:
    """Convert `qibolab.pulses.pulse` to `qibosoq.abstract.Pulse`."""

    ch = channels[ch_id]
    lo_frequency = get_lo_frequency(ch, configs)

    adc = 0  # Assign adc 0 to ensure frequency matching
    if isinstance(pulse, Acquisition):
        freq = (getattr(configs[ch_id], "frequency") - lo_frequency) / mega
        return rfsoc_pulses.Measurement(
            type="readout",
            frequency=freq,
            start_delay=start_delay,
            duration=pulse.duration * nano / micro,
            dac=0,  # Fix frequency matching to 0-dac
            adc=int(ch.path),
        )
    if isinstance(pulse, Readout):
        assert isinstance(ch, AcquisitionChannel)
        probe_id = ch.probe
        assert probe_id is not None
        adc = int(ch.path)
        ptype = "readout"
        freq = (getattr(configs[probe_id], "frequency") - lo_frequency) / mega
        amp = pulse.probe.amplitude
        rel_ph = pulse.probe.relative_phase + ch_vz_phase
        envelope = pulse.probe.envelope
    else:
        assert not isinstance(pulse, VirtualZ), (
            "VirtualZ pulse is not convertible to qibosoq.pulse (should not reach this line)"
        )
        assert not isinstance(pulse, Delay), (
            "Delay pulse is currently not convertible to qibosoq.pulse (should not reach this line)"
        )
        assert not isinstance(pulse, Align), (
            "Align pulse is currently not convertible to qibosoq.pulse (should not reach this line)"
        )

        amp = pulse.amplitude
        rel_ph = pulse.relative_phase + ch_vz_phase
        envelope = pulse.envelope
        freq = getattr(configs[ch_id], "frequency", 0)
        ptype = "drive" if freq != 0 else "flux"
        if freq != 0:
            freq = (freq - lo_frequency) / mega

    rfsoc_pulse = rfsoc_pulses.Pulse(
        frequency=freq,
        amplitude=amp,
        relative_phase=np.degrees(rel_ph),
        start_delay=start_delay,
        duration=pulse.duration * nano / micro,
        dac=int(ch.path),  # Add DAC in any case, for frequency matching
        adc=adc,
        name=str(pulse.id),
        type=ptype,
    )
    num_samples = int(sampling_rate * pulse.duration)
    return replace_pulse_shape(rfsoc_pulse, envelope, num_samples)


@convert.register
def _(par: Parameter) -> rfsoc.Parameter:
    """Convert a qibolab sweeper.Parameter into a qibosoq.Parameter."""
    return getattr(rfsoc.Parameter, par.name.upper())


@convert.register
def _(
    sweeperlist: list,  # Should be ParallelSweepers, but functools encounters an error
    sequence: PulseSequence,
    channels: dict[ChannelId, Channel],
) -> rfsoc.Sweeper:
    """Convert `qibolab.sweeper.Sweeper` to `qibosoq.abstract.Sweeper`.

    Note that any unit conversion is not done in this function (to avoid
    to do it multiple times). Conversion will be done in
    `convert_units_sweeper`.
    """

    ordered_sequence, delay_equivalence = order_pulse_sequence(sequence)
    pulse_sequence = [pulse for _, pulse, _ in ordered_sequence]

    parameters = []
    starts = []
    stops = []
    indexes = []
    expts = 0
    for sweeper in sweeperlist:
        start, stop, _ = sweeper.irange
        expts = len(sweeper.values)
        if sweeper.parameter is Parameter.offset:
            assert sweeper.channels is not None
            for ch_id in sweeper.channels:
                parameters.append(rfsoc.Parameter.BIAS)
                qubit_idx = 0
                for ch in channels:
                    if isinstance(channels[ch], DcChannel):
                        if channels[ch] == channels[ch_id]:
                            indexes.append(qubit_idx)
                        else:
                            qubit_idx += 1
                starts.append(start)
                stops.append(stop)

            if max(np.abs(starts)) > 1 or max(np.abs(stops)) > 1:
                raise ValueError(
                    "Sweeper amplitude is set to reach values higher than 1"
                )

        elif sweeper.parameter is Parameter.frequency:
            assert sweeper.channels is not None
            for ch_id in sweeper.channels:
                parameters.append(rfsoc.Parameter.FREQUENCY)
                pulse = [p for ch, p, _ in ordered_sequence if ch == ch_id][0]
                # TODO what happens if more than one pulse are on the same channel?
                indexes.append(get_index(pulse_sequence, pulse))
                starts.append(start)
                stops.append(stop)

        elif sweeper.parameter is Parameter.amplitude:
            assert sweeper.pulses is not None
            for pulse in sweeper.pulses:
                parameters.append(rfsoc.Parameter.AMPLITUDE)
                indexes.append(get_index(pulse_sequence, pulse))
                starts.append(start)
                stops.append(stop)

            if max(np.abs(starts)) > 1 or max(np.abs(stops)) > 1:
                raise ValueError(
                    "Sweeper amplitude is set to reach values higher than 1"
                )

        elif sweeper.parameter in (Parameter.phase, Parameter.relative_phase):
            assert sweeper.pulses is not None
            for pulse in sweeper.pulses:
                parameters.append(rfsoc.Parameter.RELATIVE_PHASE)
                indexes.append(get_index(pulse_sequence, pulse))
                starts.append(start)
                stops.append(stop)

        elif sweeper.parameter is Parameter.duration:
            assert sweeper.pulses is not None
            if not isinstance(sweeper.pulses[0], Delay):
                raise RuntimeError("Only delay sweepers are convertible.")

            for pulse in sweeper.pulses:
                parameters.append(rfsoc.Parameter.DELAY)
                # counting the index of the pulse (without delay pulses)
                for idx, p in enumerate(pulse_sequence):
                    if p.id in delay_equivalence:
                        if pulse.id in delay_equivalence[p.id]:
                            indexes.append(idx)

                starts.append(start)
                stops.append(stop)
        else:
            raise RuntimeError(
                f"In the sweeper conversion function, sweeper of type {type(sweeper)} is not convertible."
            )

    return rfsoc.Sweeper(
        parameters=parameters,
        indexes=indexes,
        starts=np.asarray(starts),
        stops=np.asarray(stops),
        expts=expts,
    )



"""Helper methods for (de)serializing parameters.

The format is explained in the :ref:`Loading platform parameters from
JSON <parameters_json>` example.
"""

from collections.abc import Callable, Iterable, Mapping
from typing import Annotated, Any, Optional, Union

from pydantic import (
    BeforeValidator,
    Field,
    PlainSerializer,
    TypeAdapter,
    ValidationError,
)
from pydantic_core import core_schema

from .components import (
    AcquisitionChannel,
    AcquisitionConfig,
    Channel,
    ChannelConfig,
    Config,
    DcChannel,
    DcConfig,
    IqChannel,
    IqConfig,
    IqMixerConfig,
    OscillatorConfig,
)
from .execution_parameters import ConfigUpdate, ExecutionParameters, Update
from .identifier import ChannelId, QubitId, QubitPairId
from .instruments.abstract import Instrument, InstrumentId
from .native import Native, NativeContainer, SingleQubitNatives, TwoQubitNatives
from .pulses import Acquisition, Pulse, Readout, Rectangular
from .qubits import Qubit
from .serialize import Model, replace
from .unrolling import Bounds

__all__ = [
    "ConfigKinds",
    "QubitMap",
    "InstrumentMap",
    "Hardware",
    "Parameters",
    "initialize_parameters",
]


def update_configs(configs: dict[str, Config], updates: list[ConfigUpdate]):
    """Apply updates to configs in place.

    Args:
        configs: configs to update. Maps component name to respective config.
        updates: list of config updates. Later entries in the list take precedence over earlier entries
                 (if they happen to update the same thing).
    """
    a = ConfigKinds.adapted()
    for update in updates:
        for name, changes in update.items():
            if name not in configs:
                try:
                    configs[name] = a.validate_python(changes)
                except ValidationError:
                    raise ValueError(
                        f"Cannot update configuration for unknown component {name}"
                    )
            configs[name] = replace(configs[name], **changes)


class Settings(Model):
    """Default platform execution settings."""

    nshots: int = 1000
    """Default number of repetitions when executing a pulse sequence."""
    relaxation_time: int = int(1e5)
    """Time in ns to wait for the qubit to relax to its ground state between
    shots."""

    def fill(self, options: ExecutionParameters):
        """Use default values for missing execution options."""
        if options.nshots is None:
            options = replace(options, nshots=self.nshots)

        if options.relaxation_time is None:
            options = replace(options, relaxation_time=self.relaxation_time)

        return options


class TwoQubitContainer(dict[QubitPairId, TwoQubitNatives]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        schema = handler(dict[QubitPairId, TwoQubitNatives])
        return core_schema.no_info_after_validator_function(
            cls._validate,
            schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize, info_arg=False
            ),
        )

    @classmethod
    def _validate(cls, value):
        return cls(value)

    @staticmethod
    def _serialize(value):
        return TypeAdapter(dict[QubitPairId, TwoQubitNatives]).dump_python(value)

    def __getitem__(self, key: QubitPairId):
        try:
            return super().__getitem__(key)
        except KeyError:
            value = super().__getitem__((key[1], key[0]))
            if value.symmetric:
                return value
            raise


class NativeGates(Model):
    """Native gates parameters.

    This is a container for the parameters of the whole platform.
    """

    single_qubit: dict[QubitId, SingleQubitNatives] = Field(default_factory=dict)
    coupler: dict[QubitId, SingleQubitNatives] = Field(default_factory=dict)
    two_qubit: TwoQubitContainer = Field(default_factory=dict)


ComponentId = str
"""Identifier of a generic component.

This is assumed to always be in its serialized form.
"""

# TODO: replace _UnionType with UnionType, once py3.9 will be abandoned
_UnionType = Any
_ChannelConfigT = Union[_UnionType, type[Config]]
_BUILTIN_CONFIGS: tuple[_ChannelConfigT, ...] = (ChannelConfig, Bounds)


class ConfigKinds:
    """Registered configuration kinds.

    This class is handling the known configuration kinds for deserialization.

    .. attention::

        Beware that is managing a global state. This should not be a major issue, as the
        known configurations should be fixed per run. But prefer avoiding changing them
        during a single session, unless you are clearly controlling the sequence of all
        loading operations.
    """

    _registered: list[_ChannelConfigT] = list(_BUILTIN_CONFIGS)

    @classmethod
    def extend(cls, kinds: Iterable[_ChannelConfigT]):
        """Extend the known configuration kinds.

        Nested unions are supported (i.e. :class:`Union` as elements of ``kinds``).
        """
        cls._registered.extend(kinds)

    @classmethod
    def reset(cls):
        """Reset known configuration kinds to built-ins."""
        cls._registered = list(_BUILTIN_CONFIGS)

    @classmethod
    def registered(cls) -> list[_ChannelConfigT]:
        """Retrieve registered configuration kinds."""
        return cls._registered.copy()

    @classmethod
    def adapted(cls) -> TypeAdapter:
        """Construct tailored pydantic type adapter.

        The adapter will be able to directly load all the registered
        configuration kinds as the appropriate Python objects.
        """
        return TypeAdapter(
            Annotated[
                Union[tuple(ConfigKinds._registered)], Field(discriminator="kind")
            ]
        )


def _load_configs(raw: dict[str, dict]) -> dict[ComponentId, Config]:
    a = ConfigKinds.adapted()
    return {k: a.validate_python(v) for k, v in raw.items()}


def _dump_configs(obj: dict[ComponentId, Config]) -> dict[str, dict]:
    a = ConfigKinds.adapted()
    return {k: a.dump_python(v) for k, v in obj.items()}


def _setvalue(d: dict, path: str, val: Any):
    steps = path.split(".")
    current = d
    for step in steps[:-1]:
        try:
            current = current[int(step)]
        except ValueError:
            current = current[step]

    current[steps[-1]] = val


class Parameters(Model):
    """Serializable parameters."""

    settings: Settings = Field(default_factory=Settings)
    configs: Annotated[
        dict[ComponentId, Config],
        BeforeValidator(_load_configs),
        PlainSerializer(_dump_configs),
    ] = Field(default_factory=dict)
    native_gates: NativeGates = Field(default_factory=NativeGates)

    def replace(self, update: Update) -> "Parameters":
        """Update parameters' values."""
        d = self.model_dump()
        for path, val in update.items():
            _setvalue(d, path, val)

        return self.model_validate(d)


QubitMap = Mapping[QubitId, Qubit]
InstrumentMap = Mapping[InstrumentId, Instrument]


class Hardware(Model):
    """Part of the platform that specifies the hardware configuration."""

    instruments: InstrumentMap
    qubits: QubitMap
    couplers: QubitMap = Field(default_factory=dict)


def _gate_channel(qubit: Qubit, gate: str) -> str:
    """Default channel that a native gate plays on."""
    if gate in ("RX", "RX90", "CNOT"):
        return qubit.drive
    if gate == "RX12":
        return qubit.drive_extra[(1, 2)]
    if gate == "MZ":
        return qubit.acquisition
    if gate in ("CP", "CZ", "iSWAP"):
        return qubit.flux


def _gate_sequence(qubit: Qubit, gate: str) -> Native:
    """Default sequence corresponding to a native gate."""
    channel = _gate_channel(qubit, gate)
    pulse = Pulse(duration=0, amplitude=0, envelope=Rectangular())
    if gate != "MZ":
        return Native([(channel, pulse)])

    return Native(
        [(channel, Readout(acquisition=Acquisition(duration=0), probe=pulse))]
    )


def _pair_to_qubit(pair: str, qubits: QubitMap) -> Qubit:
    """Get first qubit of a pair given in ``{q0}-{q1}`` format."""
    q = tuple(pair.split("-"))[0]
    try:
        return qubits[q]
    except KeyError:
        return qubits[int(q)]


def _native_builder(cls, qubit: Qubit, natives: set[str]) -> NativeContainer:
    """Build default native gates for a given qubit or pair.

    In case of pair, ``qubit`` is assumed to be the first qubit of the pair,
    and a default pulse is added on that qubit, because at this stage we don't
    know which qubit is the high frequency one.
    """
    return cls(
        **{
            gate: _gate_sequence(qubit, gate)
            for gate in cls.model_fields.keys() & natives
        }
    )


def _channel_config(id: ChannelId, channel: Channel) -> dict[ChannelId, Config]:
    """Default configs correspondign to a channel."""
    if isinstance(channel, DcChannel):
        return {id: DcConfig(offset=0)}
    if isinstance(channel, AcquisitionChannel):
        return {id: AcquisitionConfig(delay=0, smearing=0)}
    if isinstance(channel, IqChannel):
        configs = {id: IqConfig(frequency=0)}
        if channel.lo is not None:
            configs[channel.lo] = OscillatorConfig(frequency=0, power=0)
        if channel.mixer is not None:
            configs[channel.mixer] = IqMixerConfig(frequency=0, power=0)
        return configs
    return {id: Config()}


def initialize_parameters(
    hardware: Hardware,
    natives: Optional[set[str]] = None,
    pairs: Optional[list[str]] = None,
) -> Parameters:
    """Generates default ``Parameters`` for a given hardware configuration."""
    natives = set(natives if natives is not None else ())
    configs = {}
    for instrument in hardware.instruments.values():
        if hasattr(instrument, "channels"):
            for id, channel in instrument.channels.items():
                configs |= _channel_config(id, channel)

    single_qubit = {
        q: _native_builder(SingleQubitNatives, qubit, natives - {"CP"})
        for q, qubit in hardware.qubits.items()
    }
    coupler = {
        q: _native_builder(SingleQubitNatives, qubit, natives & {"CP"})
        for q, qubit in hardware.couplers.items()
    }
    two_qubit = {
        pair: _native_builder(
            TwoQubitNatives, _pair_to_qubit(pair, hardware.qubits), natives
        )
        for pair in (pairs if pairs is not None else ())
    }

    native_gates = NativeGates(
        single_qubit=single_qubit, coupler=coupler, two_qubit=two_qubit
    )

    return Parameters(settings=Settings(), configs=configs, native_gates=native_gates)



"""QICK-Qibosoq interface."""

import re
from collections import defaultdict
from dataclasses import asdict
from typing import cast

import numpy as np
import numpy.typing as npt
import qibosoq.components.base as rfsoc
from pydantic import Field
from qibo.config import log
from qibosoq import client
from scipy.constants import micro, nano

from qibolab._core.components.channels import AcquisitionChannel, Channel, DcChannel
from qibolab._core.components.configs import AcquisitionConfig, Config, DcConfig
from qibolab._core.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)
from qibolab._core.identifier import ChannelId, Result
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import Pulse
from qibolab._core.pulses.pulse import PulseId
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers, Parameter

from .convert import convert, convert_units_sweeper

__all__ = ["RFSoC"]


class RFSoC(Controller):
    """Instrument controlling RFSoC FPGAs."""

    bounds: str = "rfsoc/bounds"
    _sampling_rate: float = 10e9
    cfg: rfsoc.Config = Field(default_factory=rfsoc.Config)
    """Configuration dictionary required for pulse execution."""

    @property
    def sampling_rate(self):
        """Sampling rate of RFSoC."""
        return self._sampling_rate

    def connect(self):
        """Empty method to comply with Instrument interface."""

    def disconnect(self):
        """Empty method to comply with Instrument interface."""

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ) -> dict[PulseId, Result]:
        """Play a pulse sequence and retrieve feedback."""
        results = {}

        for seq in sequences:
            _validate_input_command(seq, options, sweepers)
            _update_cfg(self.cfg, options)

            fw = _firmware_loops(seq, sweepers, configs, self.channels)
            res = self._sweep(configs, seq, sweepers, len(sweepers) - fw, options, {})
            results |= _reshape_sweep_results(res, sweepers, options, fw)

        return results

    def _sweep(
        self,
        configs: dict[str, Config],
        sequence: PulseSequence,
        sweepers: list[ParallelSweepers],
        software: int,
        options: ExecutionParameters,
        updates: dict,
    ) -> dict[PulseId, Result]:
        """Execute a sweep of an arbitrary number of sweepers via recursion."""
        # If there are no software sweepers send experiment.
        # Last layer for recursion.
        if software == 0:
            if len(dict(updates)) > 0:
                log.info(f"Executing sequence with updates: {dict(updates)}")
            return self._play(configs, sequence, sweepers, options, updates)

        # use a default dictionary, merging existing values
        updates = defaultdict(dict) | ({} if updates is None else updates)

        parsweep = sweepers[0]
        results = {}
        for values in zip(*(s.values for s in parsweep)):
            # update all parallel sweepers with the respective values
            for sweeper, value in zip(parsweep, values):
                if sweeper.pulses is not None:
                    for pulse in sweeper.pulses:
                        updates[pulse.id].update({sweeper.parameter.name: value})
                if sweeper.channels is not None:
                    for channel in sweeper.channels:
                        updates[channel].update({sweeper.parameter.name: value})

            res = self._sweep(
                configs, sequence, sweepers[1:], software - 1, options, updates
            )
            results = _merge_sweep_results(results, res)

        return results

    def _play(
        self,
        configs: dict[str, Config],
        sequence: PulseSequence,
        sweepers: list[ParallelSweepers],
        options: ExecutionParameters,
        updates: dict,
    ) -> dict[PulseId, Result]:
        """Execute pulse sequence or on-hardware sweep."""
        results = {}

        self.cfg.average = (
            options.acquisition_type is not AcquisitionType.DISCRIMINATION
            and options.averaging_mode is AveragingMode.CYCLIC
        )
        opcode = (
            rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE_RAW
            if options.acquisition_type is AcquisitionType.RAW
            else rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE
        )
        toti, totq = self._execute(
            _update_configs(configs, updates),
            _update_sequence(sequence, updates),
            sweepers,
            opcode,
        )

        acq_chs = np.unique([acq[0] for acq in sequence.acquisitions])

        for idx, this_ch in enumerate(acq_chs):
            this_ch_acq = [
                (ch, acq) for ch, acq in sequence.acquisitions if ch == this_ch
            ]
            for i, q, (ch, acq) in zip(toti[idx], totq[idx], this_ch_acq):
                if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                    config = cast(AcquisitionConfig, configs[ch])
                    angle, threshold = config.iq_angle, config.threshold
                    assert angle is not None and threshold is not None
                    result = _classify_shots(np.array(i), np.array(q), angle, threshold)

                    if options.averaging_mode is AveragingMode.CYCLIC:
                        result = np.mean(result, axis=0)

                else:
                    result = np.stack([i, q], axis=-1)
                results[acq.id] = result

        return results

    def _execute(
        self,
        configs: dict[str, Config],
        sequence: PulseSequence,
        sweepers: list[ParallelSweepers],
        opcode: rfsoc.OperationCode,
    ) -> tuple[list, list]:
        """Prepare the commands dictionary to send to the qibosoq server.

        Returns lists of I and Q value measured.
        """
        converted_sweepers = [
            [convert_units_sweeper(s, self.channels, configs) for s in parsweep]
            for parsweep in sweepers
        ]
        if len(sweepers) > 0:
            if opcode == rfsoc.OperationCode.EXECUTE_PULSE_SEQUENCE_RAW:
                raise RuntimeError("Sweep not permitted in RAW mode.")
            opcode = rfsoc.OperationCode.EXECUTE_SWEEPS

        qubits = []
        for ch in self.channels:
            if isinstance(self.channels[ch], DcChannel):
                qubits.append(
                    rfsoc.Qubit(
                        bias=getattr(configs[ch], "offset", 0.0),
                        dac=int(self.channels[ch].path),
                    )
                )

        server_commands = {
            "operation_code": opcode,
            "cfg": asdict(self.cfg),
            "sequence": convert(sequence, self.sampling_rate, self.channels, configs),
            "qubits": [asdict(q) for q in qubits],
            "sweepers": [
                convert(parsweep, sequence, self.channels).serialized
                for parsweep in converted_sweepers
            ],
        }
        host, port_ = self.address.split(":")
        port = int(port_)

        try:
            return client.connect(server_commands, host, port)
            # next qibosoq version:
            # logs, (i, q) = client.connect(server_commands, host, port)
            # for board_log in logs:
            #     log.info(board_log)
            # return  (i, q)
        except RuntimeError as e:
            if "exception in readout loop" in str(e):
                log.warning(
                    "%s %s",
                    "Exception in readout loop. Attempting again",
                    "You may want to increase the relaxation time.",
                )
                return client.connect(server_commands, host, port)
            buffer_overflow = r"buffer length must be \d+ samples or less"
            if re.search(buffer_overflow, str(e)) is not None:
                log.warning("Buffer full! Use shorter pulses.")
            raise e


def _validate_input_command(
    sequence: PulseSequence,
    options: ExecutionParameters,
    sweepers: list[ParallelSweepers],
):
    """Check if sequence and execution_parameters are supported."""
    if options.acquisition_type is AcquisitionType.RAW:
        if len(sweepers) > 0:
            raise NotImplementedError(
                "Raw data acquisition is not compatible with sweepers"
            )
        if len(sequence.acquisitions) != 1:
            raise NotImplementedError(
                "Raw data acquisition is compatible only with a single readout"
            )
        if options.averaging_mode is not AveragingMode.CYCLIC:
            raise NotImplementedError("Raw data acquisition can only be averaged")
    if options.fast_reset:
        raise NotImplementedError("Fast reset is not supported")


def _update_cfg(cfg, options: ExecutionParameters):
    """Update rfsoc.Config object with new parameters."""
    if options.nshots is not None:
        cfg.reps = options.nshots
    if options.relaxation_time is not None:
        cfg.relaxation_time = options.relaxation_time * nano / micro


def _firmware_loops(
    sequence: PulseSequence,
    sweepers: list[ParallelSweepers],
    configs: dict[str, Config],
    channels: dict[ChannelId, Channel],
) -> int:
    """Check if a sweeper must be run with python loop or on hardware.

    To be run on qick internal loop a sweep must:
        * not be on the readout frequency
        * not be a duration sweeper
        * not be a duration_interpolated sweeper
        * only one pulse per channel supported
        * flux pulses are not compatible with sweepers

    Args:
        sequence (`qibolab.pulses.PulseSequence`). Pulse sequence to play.
        *sweepers (`qibosoq.abstract.Sweeper`): Sweeper objects.
    Returns:
        A boolean value true if the sweeper must be executed by python
        loop, false otherwise
    """
    if any(
        isinstance(p, Pulse) and isinstance(configs[ch], DcConfig) for ch, p in sequence
    ):
        return 0

    n = 0
    for parsweep in reversed(sweepers):
        if any(s.parameter is Parameter.duration_interpolated for s in parsweep):
            return n
        if any(s.parameter is Parameter.duration for s in parsweep):
            return n

        if any(
            s.parameter is Parameter.frequency
            and s.channels is not None
            and any(
                (isinstance(ch, AcquisitionChannel) or "probe" in ch)
                for ch in s.channels
            )
            for s in parsweep
        ):
            # if it's a sweep on the readout freq do a python sweep
            return n

        for s in parsweep:
            if s.channels is not None:
                this_channels = s.channels
            else:
                assert s.pulses is not None
                this_channels = [
                    ch for p in s.pulses for ch in sequence.pulse_channels(p.id)
                ]

            # If more than a pulse is on the sweeped channel this is a soft loop
            # What matters is the DAC number. Not the qibolab channel but the path
            for ch in this_channels:
                path = channels[ch].path
                pulses_in_path = []
                for p_ch, pulse in sequence:
                    # Type check so that ADC and DACs are not compared
                    if type(channels[p_ch]) is type(channels[ch]):
                        if channels[p_ch].path == path:
                            pulses_in_path.append(pulse)

                if len(pulses_in_path) > 1:
                    return n

        # if not disallowed, increase the amount of firmware loops
        n += 1

    return n


def _update_sequence(sequence: PulseSequence, updates: dict) -> PulseSequence:
    """Apply sweep updates to base sequence."""
    return PulseSequence(
        [(ch, e.model_copy(update=updates.get(e.id, {}))) for ch, e in sequence]
    )


def _update_configs(configs: dict[str, Config], updates: dict) -> dict[str, Config]:
    """Apply sweep updates to base configs."""
    new_dict = {k: c.model_copy(update=updates.get(k, {})) for k, c in configs.items()}
    return new_dict


def _classify_shots(
    i: npt.NDArray, q: npt.NDArray, angle: float, threshold: float
) -> npt.NDArray:
    """Classify shots through linear separation."""
    rotated = np.cos(angle) * i - np.sin(angle) * q
    return np.heaviside(np.array(rotated) - threshold, 0)


def _merge_sweep_results(
    a: dict[PulseId, Result], b: dict[PulseId, Result]
) -> dict[PulseId, Result]:
    """Merge two results dictionaries, appending common keys."""
    return {
        key: np.append(a.get(key, []), b.get(key, [])) for key in a.keys() | b.keys()
    }


def _reshape_sweep_results(
    results: dict[PulseId, Result],
    sweepers: list[ParallelSweepers],
    execution_parameters: ExecutionParameters,
    firmware_loops: int,
) -> dict[PulseId, Result]:
    """Reshape result to correct Qibolab shape."""
    if execution_parameters.acquisition_type is AcquisitionType.RAW:
        return results

    shape = [len(sweeper[0].values) for sweeper in sweepers]  # pyright: ignore

    is_not_cyclic = execution_parameters.averaging_mode is not AveragingMode.CYCLIC
    is_discrim = execution_parameters.acquisition_type is AcquisitionType.DISCRIMINATION

    if is_not_cyclic:
        shape.append(getattr(execution_parameters, "nshots", 1))

    if not is_discrim:
        shape.append(2)  # I/Q last axis
    elif is_not_cyclic and firmware_loops != 0:
        shape = [shape[-1]] + shape[:-1]

    reshaped = {}
    for key, value in results.items():
        assert value.size == np.prod(shape), (
            f"Size mismatch: value.size={value.size}, expected {np.prod(shape)}, shape={shape}"
        )

        reshaped[key] = value.reshape(shape)
        if is_not_cyclic:
            if is_discrim and firmware_loops == 0:
                reshaped[key] = np.moveaxis(reshaped[key], 0, -1)
            elif is_discrim and firmware_loops == 1:
                reshaped[key] = np.moveaxis(reshaped[key], 0, -1)
                reshaped[key] = value.reshape([shape[0], shape[2], shape[1]])
            else:
                reshaped[key] = np.moveaxis(reshaped[key], 0, -2)

    return reshaped



from collections.abc import Iterable

from pydantic import UUID4

from qibolab._core import pulses
from qibolab._core.pulses.pulse import PulseLike
from qibolab._core.serialize import Model

from .waveforms import Waveform

__all__ = []

MeasureId = UUID4
Weight = Waveform
Weights = dict[MeasureId, Weight]


class Acquisition(Model):
    num_bins: int
    index: int


Acquisitions = dict[MeasureId, Acquisition]


class AcquisitionSpec(Model):
    acquisition: Acquisition
    duration: int


def acquisitions(
    sequence: Iterable[PulseLike], num_bins: int
) -> dict[MeasureId, AcquisitionSpec]:
    return {
        acq.id: AcquisitionSpec(
            acquisition=Acquisition(num_bins=num_bins, index=i),
            duration=int(acq.duration),
        )
        for i, acq in enumerate(
            p for p in sequence if isinstance(p, (pulses.Acquisition, pulses.Readout))
        )
    }



from . import abstract, sha256, toeplitz
from .abstract import *
from .sha256 import *
from .toeplitz import *

__all__ = []
__all__ += abstract.__all__
__all__ += sha256.__all__
__all__ += toeplitz.__all__



from . import acquisition as acquisition
from . import program, sequence
from .program import *
from .sequence import *
from .sequence import compile as compile

__all__ = []
__all__ += program.__all__
__all__ += sequence.__all__



"""Qblox drivers.

https://www.qblox.com/
https://docs.qblox.com/
"""

from qibolab._core.instruments import qblox
from qibolab._core.instruments.qblox import *  # noqa: F403

__all__ = []
__all__ += qblox.__all__



from typing import Literal, Optional

from qibolab._core.serialize import Model

from ..identifiers import SlotId

__all__ = []


class PortAddress(Model):
    slot: SlotId
    ports: tuple[int, Optional[int]]
    input: bool = False

    @classmethod
    def from_path(cls, path: str):
        """Load address from :attr:`qibolab.Channel.path`."""
        els = path.split("/")
        assert len(els) == 2
        ports = els[1][1:].split("_")
        assert 1 <= len(ports) <= 2
        return cls(
            slot=int(els[0]),
            ports=(int(ports[0]), int(ports[1]) if len(ports) == 2 else None),
            input=els[1][0] == "i",
        )

    @property
    def direction(self) -> Literal["io", "out"]:
        """Signal flow direction.

        This is used for :meth:`local_address`, check its description for the in-depth
        meaning.

        .. note::

            While three directions are actually possible, there is no usage of pure
            acquisition channels in this driver (corresponding to the ``"in""``
            direction), and only support :class:`qibolab.Readout` instructions.

            For this reason, an input channel is always resolved to ``"io"``, implying
            that a same sequencer will be used to control both input (acquisitions) and
            output (probe pulses).
        """
        return "io" if self.input else "out"

    @property
    def local_address(self):
        """Physical address within the module.

        It will generate a string in the format ``<direction><channel>`` or
        ``<direction><I-channel>_<Q-channel>``.
        ``<direction>`` is ``in`` for a connection between an input and the acquisition
        path, ``out`` for a connection from the waveform generator to an output, or
        ``io`` to do both.
        The channels must be integer channel indices.
        Only one channel is present for a real mode operating sequencer; two channels
        are used for complex mode.

        .. note::

            Description adapted from
            https://docs.qblox.com/en/main/api_reference/cluster.html#qblox_instruments.Cluster.connect_sequencer
        """
        for port in self.ports:
            if port is not None:
                assert port > 0
        channels = (
            str(self.ports[0] - 1)
            if self.ports[1] is None
            else f"{self.ports[0] - 1}_{self.ports[1] - 1}"
        )
        return f"{self.direction}{channels}"



from typing import Annotated, Literal, Union

import numpy as np
from pydantic import Field

from ..serialize import Model

__all__ = ["ExponentialFilter", "FiniteImpulseResponseFilter"]


class ExponentialFilter(Model):
    """Exponential filter.

    Correct filters for signal with behavior
    1 + amplitude * e^(-t/tau).
    """

    kind: Literal["exp"] = "exp"
    amplitude: float
    """Amplitude for exponential term."""
    tau: float
    """Time decay in exponential in samples unit."""

    def _compute_filter(self) -> tuple[list[float], list[float]]:
        """Computing feedback and feedforward taps."""
        alpha = 1 - np.exp(-1 / (self.tau * (1 + self.amplitude)))
        k = (
            self.amplitude / ((1 + self.amplitude) * (1 - alpha))
            if self.amplitude < 0
            else self.amplitude / (1 + self.amplitude - alpha)
        )
        b0 = 1 - k + k * alpha
        b1 = -(1 - k) * (1 - alpha)
        a0 = 1
        a1 = -(1 - alpha)
        return [a0, a1], [b0, b1]

    @property
    def feedback(self) -> list[float]:
        return self._compute_filter()[0]

    @property
    def feedforward(self) -> list[float]:
        return self._compute_filter()[1]


class FiniteImpulseResponseFilter(Model):
    """FIR filter."""

    kind: Literal["fir"] = "fir"
    coefficients: list[float]

    @property
    def feedforward(self):
        return self.coefficients


Filter = Annotated[
    Union[ExponentialFilter, FiniteImpulseResponseFilter],
    Field(discriminator="kind"),
]



from qibolab._core.identifier import ChannelId
from qibolab._core.instruments.qblox.sequence.acquisition import Acquisition, Weight
from qibolab._core.instruments.qblox.sequence.sequence import Q1Sequence
from qibolab._core.sequence import PulseSequence

from .sequence.waveforms import Waveform

__all__ = []

WAVEFORM_MEMORY = 2**14
"""Maximum waveform memory available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#waveform-memory
"""
WAVEFORM_NUMBER = 2**10
"""Maximum number of waveforms available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#waveform-memory
"""

WEIGHT_MEMORY = 2**14
"""Maximum waveform memory available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#integrator
"""
WEIGHT_NUMBER = 2**5
"""Maximum number of weights available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#weight-memory
"""

ACQUISITION_MEMORY = 2**17
"""Maximum acquisition memory available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#binning-and-averaging
"""

ACQUISITION_NUMBER = 2**5
"""Maximum number of acquisitions available.

https://docs.qblox.com/en/main/cluster/q1_sequence_processor.html#:~:text=Square%20Weight%20Acquisitions
"""


def assert_channels_exclusion(ps: PulseSequence, excluded: set[ChannelId]) -> None:
    """Deny group of channels.

    Mainly used to deny probe channels, since probe events are currently required to be
    bundled with acquisitions, cf. :class:`qibolab.Readout`.

    An extended discussion about the probe-acquisition merge is found at:
    https://github.com/qiboteam/qibolab/pull/1088#issuecomment-2637800857
    """
    assert not any(ch in excluded for ch, _ in ps), (
        "Probe channels can not be controlled independently, please use Readout events."
    )


def assert_waveform_memory(waveforms: list[Waveform]) -> None:
    """Assert waveform memory limitations."""
    assert len(waveforms) <= WAVEFORM_NUMBER
    assert sum(len(w.data) for w in waveforms) <= WAVEFORM_MEMORY


def assert_weight_memory(weights: list[Weight]) -> None:
    """Assert weight memory limitations."""
    assert len(weights) <= WEIGHT_NUMBER
    assert sum(len(w.data) for w in weights) <= WEIGHT_MEMORY


def assert_acquisition_memory(acquisitions: list[Acquisition]) -> None:
    """Assert acquisition memory limitations."""
    assert len(acquisitions) <= ACQUISITION_NUMBER
    assert sum(a.num_bins for a in acquisitions) <= ACQUISITION_MEMORY


def validate_sequence(sequence: Q1Sequence) -> None:
    """Validate sequence elements."""
    assert_acquisition_memory(list(sequence.acquisitions.values()))
    assert_waveform_memory(list(sequence.waveforms.values()))
    assert_weight_memory(list(sequence.weights.values()))



import inspect
import re
import shutil
import textwrap
from typing import Annotated, Any, Iterable, Optional, Sequence, Union

from pydantic import (
    AfterValidator,
    BeforeValidator,
    ConfigDict,
    GetCoreSchemaHandler,
    computed_field,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic_core import core_schema

from ....serialize import Model

__all__ = ["Program"]


class Register(Model):
    number: int

    @model_validator(mode="before")
    @classmethod
    def load(cls, data: Any) -> Any:
        if isinstance(data, str):
            try:
                assert data[0] == "R"
            except TypeError:
                raise ValueError("Register representation is not a string")
            num = int(data[1:])
        elif isinstance(data, dict):
            num = data["number"]
        else:
            raise ValueError(f"Register not recognized '{data}'")

        assert 0 <= num < 64
        return {"number": num}

    @model_serializer
    def dump(self) -> str:
        return f"R{self.number}"


class Reference(Model):
    label: str

    @model_validator(mode="before")
    @classmethod
    def load(cls, data: Any) -> Any:
        if isinstance(data, str):
            assert data[0] == "@"
            return {"label": data[1:]}
        if isinstance(data, dict):
            return data
        raise ValueError(f"Reference not recognized '{data}'")

    @model_serializer
    def dump(self) -> str:
        return f"@{self.label}"


def _value_bounds(n: int):
    if abs(n) >= 2**32:
        raise ValueError("Register value out of bounds")
    return n


MultiBaseInt = Annotated[
    int,
    BeforeValidator(lambda n: int(n, 0) if isinstance(n, str) else n),
    BeforeValidator(lambda n: _value_bounds(n) if isinstance(n, int) else n),
]
Immediate = Union[MultiBaseInt, Reference]
Value = Union[Register, Immediate]

CAMEL_TO_SNAKE = re.compile("(?<=[a-z0-9])(?=[A-Z])(?!^)(?=[A-Z][a-z])")


class Instr(Model):
    model_config = ConfigDict(extra="ignore")

    @model_validator(mode="before")
    @classmethod
    def load(cls, data: Any) -> Any:
        """Leverage automated tagging to resolve conflicts.

        This is required to unambiguously deserialize the instruction. Cf.
        :meth:`instr`.

        It is kind of surrogate of Pyantic's discriminated unions, without the need of
        manually labeling each class, since the label is derived from the type.
        """
        if "instr" in data:
            assert data["instr"] == cls.keyword()
        return data

    @computed_field
    @property
    def instr(self) -> str:
        """Store instruction information in the serialization.

        The instruction is non-uniquely identified by the attributes
        structure, thus logging its identity is required, but the
        information at runtime is only contained in the model's type.
        """
        return self.keyword()

    @classmethod
    def keyword(cls) -> str:
        return CAMEL_TO_SNAKE.sub("_", cls.__name__).lower()

    @classmethod
    def from_args(cls, *args):
        return cls(**dict(zip(cls.model_fields.keys(), args)))

    @property
    def args(self) -> list:
        return list(
            {k: v for k, v in self.model_dump().items() if k != "instr"}.values()
        )

    def asm(self, key_width: Optional[int] = None) -> str:
        key = self.keyword()
        if key_width is None:
            key_width = len(key)
        instr = f"{key:<{key_width + 1}}"
        return (instr + ",".join([str(a) for a in self.args])).strip()


class Illegal(Instr):
    """Illegal instruction.

    Instruction that should not be executed. If it is executed, the
    sequencer will stop with the illegal instruction flag set.
    """


class Stop(Instr):
    """Stop the sequencer.

    Instruction that stops the sequencer.
    """


class Nop(Instr):
    """No operation.

    No operation instruction, that does nothing. It is used to pass a
    single cycle in the classic part of the sequencer without any
    operations.
    """


Control = Union[Illegal, Stop, Nop]
"""Control instructions."""


class Jmp(Instr):
    """Jump.

    Jump to the instruction indicated by :attr:`address`.
    """

    address: Value


class Jge(Instr):
    """Jump if greater.

    If :attr:`a` is greater than or equal to :attr:`b`, jump to the instruction
    indicated by :attr:`address`.
    """

    a: Register
    b: Immediate
    address: Value


class Jlt(Instr):
    """Jump if smaller.

    If :attr:`a` is less than :attr:`b`, jump to the instruction indicated by
    :attr:`address`.
    """

    a: Register
    b: Immediate
    address: Value


class Loop(Instr):
    """Loop.

    Decrement :attr:`a` by one and, if the result is non-zero, jump to the instruction
    indicated by :attr:`address`.
    """

    a: Register
    address: Value


Jump = Union[Jmp, Jge, Jlt, Loop]
"""Jump instructions."""


class Move(Instr):
    """Move value.

    :attr:`source` is moved / copied to :attr:`destination`.

    ::

        destination = source
    """

    source: Value
    destination: Register


class Not(Instr):
    """Bit-wise inversion.

    Bit-wise invert :attr:`source` and move the result to :attr:`destination`.

    ::

        destination = ~source
    """

    source: Value
    destination: Register


class Add(Instr):
    """Addition.

    Add :attr:`b` to :attr:`a` and move the result to :attr:`destination`.

    ::

        destination = a + b
    """

    a: Register
    b: Value
    destination: Register


class Sub(Instr):
    """Subtraction.

    Subtract :attr:`b` from :attr:`a` and move the result to :attr:`destination`.

    ::

        destination = a - b
    """

    a: Register
    b: Value
    destination: Register


class And(Instr):
    """Bit-wise conjuction.

    Bit-wise AND :attr:`a` and :attr:`b` and move the result to :attr:`destination`.

    ::

        destination = a & b
    """

    a: Register
    b: Value
    destination: Register


class Or(Instr):
    """Bit-wise disjuction.

    Bit-wise OR :attr:`a` and :attr:`b` and move the result to :attr:`destination`.

    ::

        destination = a | b
    """

    a: Register
    b: Value
    destination: Register


class Xor(Instr):
    """Bit-wise exclusive disjuction.

    Bit-wise XOR :attr:`a` and :attr:`b` and move the result to :attr:`destination`.

    ::

        destination = a ^ b
    """

    a: Register
    b: Value
    destination: Register


class Asl(Instr):
    """Bit-wise left-shift.

    Bit-wise left-shift :attr:`a` by :attr:`b` number of bits and move the result to
    :attr:`destination`.

    ::

        destination = a << b
    """

    a: Register
    b: Value
    destination: Register


class Asr(Instr):
    """Bit-wise right-shift.

    Bit-wise right-shift :attr:`a` by :attr:`b` number of bits and move the result to
    :attr:`destination`.

    ::

        destination = a << b
    """

    a: Register
    b: Value
    destination: Register


Arithmetic = Union[Move, Not, Add, Sub, And, Or, Xor, Asl, Asr]
"""Arithmetic instructions."""


class SetMrk(Instr):
    """Set marker.

    Set marker output channels to ``val`` (bits 0-3), where the bit index
    corresponds to the channel index for baseband modules.

    For QCM-RF modules, bit indices 0 & 1 correspond to enabling output
    1 and 2 switches respectively; indices 2 & 3 correspond to marker
    output 2 and 1 (sic!) respectively.

    For QRM-RF modules, bit index 0 is inactive, bit index 1 corresponds
    to enabling output 1 switch; indices 2 & 3 correspond to marker
    output 1 and 2 (sic!) respectively.

    The values are OR'ed by that of other sequencers.
    """

    mask: Value


class SetFreq(Instr):
    """Set frequency.

    Set the frequency of the NCO used by the AWG and acquisition using
    :attr:`value`.
    The frequency is divided into 4e9 steps between -500 and 500 MHz and
    expressed as an integer between -2e9 and 2e9 (e.g.1 MHz=4e6).
    """

    value: Value


class ResetPh(Instr):
    """Phase reset.

    Reset the absolute phase of the NCO used by the AWG and acquisition
    to 0°. This also resets any relative phase offsets that were already
    statically or dynamically set.
    """


class SetPh(Instr):
    """Set absolute phase.

    Set the absolute phase of the NCO used by the AWG and acquisition
    using value. The phase is divided into 1e9 steps between 0° and
    360°, expressed as an integer between 0 and 1e9 (e.g 45°=125e6).
    """

    value: Value


class SetPhDelta(Instr):
    """Set phase offset.

    Set an offset on top of the relative phase of the NCO used by the
    AWG and acquisition. The offset is applied on top of the phase set
    using ``set_ph``. See :class:`SetPh` for more details regarding the
    argument.
    """

    value: Value


class SetAwgGain(Instr):
    """Set AWG gain.

    Set AWG gain for path 0 using :attr:`value_0` and path 1 using :attr:`value_1`.
    Both are integers in ``[-32 768, 32 767]``.
    """

    value_0: Value
    value_1: Value

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.value_0) is type(self.value_1)
        return self

    @field_validator("value_0", "value_1")
    @classmethod
    def check_range(cls, v: Value) -> Value:
        if isinstance(v, int):
            assert -32_768 <= v <= 32_767
        return v


class SetAwgOffs(Instr):
    """Set AWG offset.

    Set AWG offset for path 0 using :attr:`value_0` and path 1 using :attr:`value_1`.
    Both are integers in ``[-32 768, 32 767]``.
    """

    value_0: Value
    value_1: Value

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.value_0) is type(self.value_1)
        return self

    @field_validator("value_0", "value_1")
    @classmethod
    def check_range(cls, v: Value) -> Value:
        if isinstance(v, int):
            assert -32_768 <= v <= 32_767
        return v


ParamOps = Union[SetMrk, SetFreq, ResetPh, SetPh, SetPhDelta, SetAwgGain, SetAwgOffs]
"""Parameter operations.

The parameters are latched and only updated when the ``upd_param``, ``play``, ``acquire``,
``acquire_weighed`` or ``acquire_ttl`` instructions are executed.
"""

Q1Instr = Union[Control, Jump, Arithmetic, ParamOps]
"""Q1 Instructions.

These instructions are used to compose and manipulate the arguments of
real-time instructions. They are always executed before the next real-
time instruction, and therefore take zero wall-time.
"""


class SetCond(Instr):
    """Condition all subsequent instructions.

    Enable/disable conditionality on all following real-time
    instructions based on :attr:`enable`. The condition is based on the
    trigger network address counters being thresholded based on the associated
    counter threshold parameters set through QCoDeS.

    The results are masked using :attr:`mask` (bits 0-14), where the bit
    index plus one corresponds to the trigger address.

    This creates a selection to include in the final logical operation
    set using :attr:`operator`. Logical operators are OR, NOR, AND, NAND,
    XOR, XNOR, where a value for operator of 0 is OR and 5 is XNOR
    respectively.

    The logical operation result (true/false) determines the condition.
    If the condition is true upon evaluation, the next real-time
    instruction is executed. Else the real-time path ignores the
    instruction and waits for :attr:`else_duration` nanoseconds before
    continuing to the next.

    All following real-time instructions are subject to the same
    condition, until either the conditionality is disabled or updated.
    Disabling the conditionality does not affect the address counters
    and does not trigger else_duration.
    """

    enable: Value
    mask: Value
    operator: Value
    else_duration: Immediate

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.enable) is type(self.mask)
        assert type(self.enable) is type(self.operator)
        return self


Conditional = SetCond
"""Conditional instructions."""


class UpdParam(Instr):
    """Update parameters.

    Update the latched parameters, and then wait for :attr:`duration` nanoseconds.
    """

    duration: Immediate


class Play(Instr):
    """Play waveforms.

    Update the latched parameters, interrupt currently playing waves and
    start playing AWG waveforms stored at indexes :attr:`wave_0` on path 0 and
    :attr:`wave_1` on path 1.
    """

    wave_0: Value
    wave_1: Value
    duration: Immediate

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.wave_0) is type(self.wave_1)
        return self


class Acquire(Instr):
    """Start new acquisition.

    Update the latched parameters, interrupt currently active
    acquisitions and start the acquisition referred to using index
    :attr:`acquisition` and store the bin data in :attr:`bin` index bin.

    Integration is executed using a square weight with a preset length
    through the associated QCoDeS parameter.
    """

    acquisition: Immediate
    bin: Value
    duration: Immediate


class AcquireWeighed(Instr):
    """Start new acquisition with given weights.

    Update the latched parameters, interrupt currently active
    acquisitions and start the acquisition referred to using index
    :attr:`acquisition` and store the bin data in :attr:`bin` index bin.

    Integration is executed using weights stored at indices :attr:`weight_0` for
    path 0 and :attr:`weight_1` for path 1.
    """

    acquisition: Immediate
    bin: Value
    weight_0: Value
    weight_1: Value
    duration: Immediate

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.bin) is type(self.weight_0)
        assert type(self.bin) is type(self.weight_1)
        return self


class AcquireTtl(Instr):
    """Start TTL trigger acquisition.

    Update the latched parameters, start the TTL trigger acquisition
    referred to using index :attr:`acquisition` and store the bin data in bin
    index :attr:`bin`. Enable the acquisition by writing 1 to :attr:`enable`.

    The TTL trigger acquisition has to be actively disabled afterwards
    by writing 0 to :attr:`enable`.
    """

    acquisition: Immediate
    bin: Value
    enable: Immediate
    duration: Immediate


Io = Union[UpdParam, Play, Acquire, AcquireWeighed, AcquireTtl]
"""Real-time IO operation instructions.

The execution of any of these instructions will cause the latched
parameters to be updated.
"""


class SetLatchEn(Instr):
    """Toggle trigger counters.

    Enable/disable all trigger network address counters. Once enabled,
    the trigger network address counters will count all triggers on the
    trigger network. When disabled, the counters hold their last values.
    """

    enable: Value
    duration: Immediate


class LatchRst(Instr):
    """Reset trigger counters.

    Reset all trigger network address counters back to 0.
    """

    duration: Value


Trigger = Union[SetLatchEn, LatchRst]
"""Real-time trigger count control instructions."""


class Wait(Instr):
    """Wait.

    Wait for :attr:`duration` nanoseconds.
    """

    duration: Value


class WaitTrigger(Instr):
    """Wait trigger.

    Wait for a trigger on the trigger network at the address set using :attr:`trigger`.
    """

    trigger: Value
    duration: Value

    @model_validator(mode="after")
    def check_signature(self):
        assert type(self.trigger) is type(self.duration)
        return self


class WaitSync(Instr):
    """Syncronization.

    Wait for SYNQ to complete on all connected sequencers over all
    connected instruments.
    """

    duration: Value


WaitOps = Union[Wait, WaitTrigger, WaitSync]
"""Real-time wait operation instructions."""

RealTimeInstr = Union[Conditional, Io, Trigger, WaitOps]
"""Real-time instructions.

These instructions have a duration argument, which corresponds to the
wall-time of an experiment. The duration is counted from the beginning
of an instruction. This is what enables e.g. parallel playback and
acquisition. The duration can be set at a resolution of 1 ns, with a
minimum of 4 ns.

The real-time instructions to the pipeline form a time-line of real-time
operations. To ensure deterministic behavior, this time-line cannot be
interrupted/broken. As a result, once the real-time pipeline is started
by the first real-time instruction in a sequence, it cannot be stalled.
Therefore, the real-time pipeline will never wait for the Q1 processor.
In case of a conflict, the sequencer will halt and raise an error flag.
"""

Instruction = Union[Q1Instr, RealTimeInstr]


INSTRUCTIONS = {
    c.keyword(): c
    for c in locals().values()
    if inspect.isclass(c) and issubclass(c, Instr)
}


def _format_comment(text: str, width: Optional[int] = None) -> str:
    lines = [""]
    for block in text.splitlines():
        lines.extend(
            textwrap.wrap(
                block, width=width - 2, break_long_words=False, break_on_hyphens=False
            )
            if width is not None
            else [block]
        )
    return "\n# ".join(lines)[1:]


class Comment(str):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def asm(self, width: Optional[int] = None) -> str:
        return _format_comment(self, width) + "\n"


class Line(Model):
    instruction: Instruction
    label: Optional[str] = None
    comment: Optional[Annotated[str, AfterValidator(lambda c: c.strip())]] = None

    def __rich_repr__(self):
        yield self.instruction
        yield "label", self.label, None
        yield "comment", self.comment, None

    @classmethod
    def instr(cls, instruction: Instruction):
        """Shortcut for simple lines instantiation."""
        return cls(instruction=instruction)

    def update(self, update: dict[str, Any]) -> "Line":
        """Update parameters in the underlying instructions keeping the same comments."""
        new_instr = self.instruction.model_copy(update=update)
        return self.model_copy(update={"instruction": new_instr})

    def asm(
        self,
        width: Optional[int] = None,
        label_width: Optional[int] = None,
        instr_name_width: Optional[int] = None,
        instr_width: Optional[int] = None,
    ) -> str:
        if label_width is None:
            label_width = len(self.label) if self.label is not None else -2
        label = f"{self.label}:" if self.label is not None else ""
        if instr_name_width is None:
            instr_name_width = len(self.instruction.keyword())
        if instr_width is None:
            instr_width = len(self.instruction.asm(instr_name_width))
        code = f"{label:<{label_width + 2}}{self.instruction.asm(instr_name_width):<{instr_width + 1}}"
        if self.comment is None:
            return code
        comment = _format_comment(
            self.comment, width - len(code) if width is not None else None
        ).splitlines()
        if len(comment) == 0:
            return code
        return "\n".join(
            [code + comment[0]] + [" " * len(code) + c for c in comment[1:]]
        )


Element = Union[Line, Comment]
Lineable = Union[Line, Instruction]
Block = Sequence[Lineable]
BlockList = list[Lineable]
BlockIter = Iterable[Lineable]


class Program(Model):
    elements: list[Element]

    @classmethod
    def from_elements(cls, elements: list[Element]):
        comments = []
        elements_ = []

        # group comments
        for el in elements:
            if isinstance(el, Comment):
                comments.append(el)
                continue
            if len(comments) > 0:
                comment = "\n".join(comments)
                elements_.append(Comment(comment))
                comments = []
            elements_.append(el)

        return cls(elements=elements_)

    def asm(self, width: Optional[int] = None, comments: bool = True) -> str:
        max_label_len = max(
            (
                len(line.label)
                for line in self.elements
                if isinstance(line, Line) and line.label is not None
            ),
            default=None,
        )
        max_instr_name_len = max(
            (
                len(line.instruction.keyword())
                for line in self.elements
                if isinstance(line, Line)
            ),
            default=None,
        )
        max_instr_len = max(
            (
                len(line.instruction.asm(max_instr_name_len))
                for line in self.elements
                if isinstance(line, Line)
            ),
            default=None,
        )
        code = "\n".join(
            (
                (el if comments else el.model_copy(update={"comment": None})).asm(
                    width,
                    label_width=max_label_len,
                    instr_name_width=max_instr_name_len,
                    instr_width=max_instr_len,
                )
                if isinstance(el, Line)
                else (el.asm(width) if comments else "")
            )
            for el in self.elements
        )
        return (
            code if comments else re.sub("^\n*", "", re.sub("\n+", "\n", code))
        ) + "\n"

    def __rich__(self):
        return self.asm(width=shutil.get_terminal_size((None, None)).columns)



"""Abstract engine for platform emulation."""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Callable, Protocol, Union

from ....serialize import Model

__all__ = [
    "SimulationEngine",
    "Operator",
    "TimeDependentOperator",
    "OperatorEvolution",
]


class Operator(Protocol):
    """Abstract operator interface.

    It can represents both operators and states in the Hilbert space.
    """

    n: int
    """Dimension of Hilbert space."""

    def dag(self) -> "Operator":
        """Return the adjoint of the operator."""

    def __add__(self, other: "Operator") -> "Operator":
        """Add two operators."""


class TimeDependentOperator(Protocol):
    """Abstract time dependent operator interface."""

    operator: Operator
    """Operator."""
    time: Callable[[float, dict], float]
    """Time function."""


class EvolutionResult(Protocol):
    """Result from evolution."""

    states: list[Operator]
    """List of Operators."""


@dataclass
class OperatorEvolution:
    """Abstract operator evolution interface."""

    operators: list[Union[Operator, TimeDependentOperator]] = field(
        default_factory=list
    )


class SimulationEngine(Model, ABC):
    """Parent class for generic simulation engine."""

    @property
    @abstractmethod
    def engine(self):
        """Engine module."""

    @abstractmethod
    def evolve(
        self,
        hamiltonian: Operator,
        initial_state: Operator,
        time: list[float],
        collapse_operators: list[Operator] = None,
        **kwargs,
    ) -> EvolutionResult:
        """Evolve the system."""

    @abstractmethod
    def create(self, n: int) -> Operator:
        """Create operator for n levels system."""

    @abstractmethod
    def destroy(self, n: int) -> Operator:
        """Destroy operator for n levels system."""

    @abstractmethod
    def identity(self, n: int) -> Operator:
        """Identity operator for n levels system."""

    @abstractmethod
    def tensor(self, operators: list[Operator]) -> Operator:
        """Tensor product of a list of operators."""

    @abstractmethod
    def expand(
        self, op: Operator, targets: Union[int, list[int]], dims: list[int]
    ) -> Operator:
        """Expand operator in larger Hilbert space."""

    @abstractmethod
    def basis(self, n: int, state: int) -> Operator:
        """Basis operator for n levels system."""



import json
from typing import Any, Optional

from qblox_instruments import SequencerStatus, SequencerStatuses

__all__ = []


class MockSequencer:
    def __init__(self, idx: int, ancestors: list) -> None:
        self.idx = idx
        self.register = {"calls": []}
        self.ancestors = [self] + ancestors

    @property
    def seq_idx(self):
        return self.idx

    def __getattribute__(self, name: str):
        if name in ["idx", "register", "ancestors", "seq_idx"]:
            return super().__getattribute__(name)

        log: dict["str", Any] = {"name": name}
        self.register["calls"].append(log)

        def wrapped(*args, **kwargs):
            log["args"], log["kwargs"] = args, kwargs

        return wrapped


class MockModule:
    def __init__(self, slot: int) -> None:
        self.slot_idx = slot
        self.register = {"calls": []}
        self.sequencers = [MockSequencer(i, [self]) for i in range(6)]

    def present(self) -> bool:
        return True

    @property
    def is_rf_type(self) -> bool:
        return True

    @property
    def is_qrm_type(self) -> bool:
        return True

    def snapshot(self) -> dict:
        return self.register | {
            "sequencers": {seq.idx: seq.register for seq in self.sequencers}
        }

    def __getattribute__(self, name: str):
        if name in [
            "slot_idx",
            "register",
            "sequencers",
            "present",
            "is_rf_type",
            "is_qrm_type",
            "snapshot",
        ]:
            return super().__getattribute__(name)

        log: dict["str", Any] = {"name": name}
        self.register["calls"].append(log)

        def wrapped(*args, **kwargs):
            log["args"], log["kwargs"] = args, kwargs

        return wrapped


class MockCluster:
    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs
        self.resets: int = 0
        self.modules = [MockModule(slot) for slot in range(21)]

    def reset(self) -> None:
        self.resets += 1

    def close(self) -> None:
        pass

    @property
    def records(self) -> dict:
        return {
            "args": self.args,
            "kwargs": self.kwargs,
            "resets": self.resets,
            "modules": {mod.slot_idx: mod.snapshot() for mod in self.modules},
        }

    def snapshot(self) -> str:
        return json.dumps(self.records)

    @property
    def sequences(self) -> dict[tuple[int, int], Optional[dict]]:
        return {
            k: v
            for k, v in {
                (mod.slot_idx, seq.idx): next(
                    iter(
                        call["args"][1]
                        for call in seq.register["calls"]
                        if call["name"] == "set" and call["args"][0] == "sequence"
                    ),
                    None,
                )
                for mod in self.modules
                for seq in mod.sequencers
            }.items()
            if v is not None
        }

    @property
    def programs(self) -> dict[tuple[int, int], Optional[dict]]:
        return {
            id_: seq["program"] if seq is not None else None
            for id_, seq in self.sequences.items()
        }

    def get_sequencer_status(
        self, slot: int, sequencer: int, *, timeout: int
    ) -> SequencerStatus:
        return SequencerStatus(
            status=SequencerStatuses.OKAY,
            state=None,
            info_flags=None,
            warn_flags=None,
            err_flags=None,
            log=None,
        )

    def get_acquisition_status(self, slot: int, sequencer: int, *, timeout: int):
        pass

    def get_acquisitions(self, slot: int, sequencer: int) -> dict:
        return {}

    def reference_source(self, value=None) -> None:
        pass



from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
from serial import Serial

from ..abstract import Instrument
from .extractors import Extractor, ShaExtractor

__all__ = ["QRNG"]


def read(port: Serial, n: int = 1, nbytes: int = 4) -> List[int]:
    """Read raw samples from the QRNG device serial output.

    In the entropy mode of the device, these typically follow a
    normal distribution.

    Args:
        n: Number of samples to retrieve.
        nbytes: Number of bytes to read from serial port to generate one raw sample.
    """
    samples = []
    while len(samples) < n:
        num_str = ""
        while len(num_str) < nbytes:
            sample = port.read(1)
            if sample == b" ":
                break
            num_str += sample.decode("utf-8")
        try:
            samples.append(int(num_str))
        except ValueError:
            pass
    return samples


class QRNG(Instrument):
    """Driver to sample numbers from a Quantum Random Number Generator (QRNG).

    See :ref:`qrng` for example usage.
    """

    address: str
    baudrate: int = 115200
    extractor: Extractor = ShaExtractor()
    port: Optional[Serial] = None

    bytes_per_number: int = 4
    """Number of bytes to read from serial port to generate one raw sample."""

    def connect(self):
        if self.port is None:
            self.port = Serial(self.address, self.baudrate)

    def disconnect(self):
        if self.port is not None:
            self.port.close()
            self.port = None

    def read(self, n: int) -> List[int]:
        """Read raw samples from the QRNG device serial output.

        In the entropy mode of the device, these typically follow a
        normal distribution.

        Args:
            n: Number of samples to retrieve.
        """
        return read(self.port, n, self.bytes_per_number)

    def random(self, size: Optional[Union[int, Sequence[int]]] = None) -> npt.NDArray:
        """Returns random floats following uniform distribution in [0, 1].

        Args:
            size: Shape of the returned array (to behave similarly to ``np.random.random``).
        """
        n = np.prod(size)
        nraw = self.extractor.num_raw_samples(n)
        raw = self.read(nraw)
        extracted = self.extractor.extract(raw)[:n]
        return np.reshape(extracted, size)



from collections.abc import Iterable, Sequence
from typing import Optional

from qibolab._core.execution_parameters import AveragingMode, ExecutionParameters
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses.pulse import PulseId, PulseLike
from qibolab._core.sweeper import ParallelSweepers

from ..q1asm.ast_ import (
    Block,
    Instruction,
    Line,
    Move,
    Nop,
    Program,
    Stop,
    Wait,
)
from .acquisition import AcquisitionSpec, MeasureId
from .experiment import experiment
from .loops import LoopSpec, Registers, loop, loops
from .sweepers import Param, params, params_reshape, sweep_sequence, update_instructions
from .transpile import transpile
from .waveforms import WaveformIndices

__all__ = ["Program"]


def setup(
    loops: Sequence[LoopSpec],
    params: list[Param],
    channel: set[ChannelId],
    pulses: set[PulseId],
) -> Block:
    """Build preparation phase. Ending with synchronization.

    This will set up all the registers used for iterations, parameters
    updates (sweeps), and to track the bin index where to save the
    results.

    The initial values for channel-wide parameters is also set.

    It is guaranteed that the final instruction of this block will
    trigger the synchronization among modules, with no other instruction
    trailing.
    """
    return (
        [
            Line(
                instruction=Move(source=0, destination=Registers.bin.value),
                comment="init bin counter",
            ),
            Line(
                instruction=Move(source=0, destination=Registers.bin_reset.value),
                comment="init bin reset",
            ),
        ]
        + [
            Line(
                instruction=Move(source=lp.length, destination=lp.reg),
                comment=f"init {lp.description} counter",
            )
            for lp in loops
        ]
        + [
            Line(
                instruction=Move(source=p.start, destination=p.reg),
                comment=f"init {p.description}",
            )
            for p in params
            if p.channel in channel or p.pulse in pulses
        ]
        # wait one clock cycle before parameters' update
        # cf. .loops._sweep_update()
        + [Nop()]
        + [
            inst
            for p in params
            if p.channel in channel
            for inst in update_instructions(p.role, p.reg)
        ]
    )


def finalization() -> list[Instruction]:
    """Finalize.

    Currently only stopping the sequencer.
    """
    return [Stop()]


def program(
    sequence: Iterable[PulseLike],
    waveforms: WaveformIndices,
    acquisitions: dict[MeasureId, AcquisitionSpec],
    options: ExecutionParameters,
    sweepers: list[ParallelSweepers],
    channel: set[ChannelId],
    time_of_flight: Optional[float],
    padding: int,
) -> Program:
    """Generate sequencer program."""
    assert options.nshots is not None
    assert options.relaxation_time is not None

    loops_ = loops(
        sweepers,
        options.nshots,
        inner_shots=options.averaging_mode is AveragingMode.SEQUENTIAL,
    )
    params_ = params(sweepers, allocated=max(lp.reg.number for lp in loops_))
    indexed_params = params_reshape(params_)
    sweepseq = sweep_sequence(
        sequence, [p for v in indexed_params.values() for p in v.pulse]
    )
    experiment_ = [
        *experiment(sweepseq, waveforms, acquisitions, time_of_flight),
        # add 4 spare ns to ensure minimum duration
        Wait(duration=padding + 4),
    ]
    singleshot = options.averaging_mode is AveragingMode.SINGLESHOT
    pulses = {p[0].id for p in sweepseq}

    return transpile(
        [
            el
            for block in [
                setup(loops_, params_, channel, pulses),
                loop(
                    experiment_,
                    loops_,
                    indexed_params,
                    options.relaxation_time,
                    singleshot,
                    channel,
                    pulses,
                ),
                finalization(),
            ]
            for el in block
        ]
    )



from dataclasses import dataclass, field
from typing import Literal, Union

from qibolab._core.components import OscillatorConfig

from ..components import (
    MwFemOscillatorConfig,
    OctaveOscillatorConfig,
    QmAcquisitionConfig,
)
from ..components.configs import OctaveOutputModes

__all__ = [
    "ModuleTypes",
    "MwFemOutput",
    "MwFemInput",
    "OctaveOutput",
    "OctaveInput",
    "Controller",
    "Octave",
    "ControllerId",
    "Controllers",
]


@dataclass(frozen=True)
class AnalogInput:
    offset: float = 0.0
    gain_db: int = 0

    @classmethod
    def from_config(cls, config: QmAcquisitionConfig):
        return cls(offset=config.offset, gain_db=config.gain)


@dataclass
class MwFemOutput:
    upconverters: dict[int, dict[Literal["frequency"], float]]
    band: int
    sampling_rate: float
    full_scale_power_dbm: int

    @classmethod
    def from_config(cls, config: MwFemOscillatorConfig):
        upconverters = {config.upconverter: {"frequency": config.frequency}}
        return cls(
            upconverters=upconverters,
            band=config.band,
            sampling_rate=config.sampling_rate,
            full_scale_power_dbm=config.power,
        )

    def update(self, config: MwFemOscillatorConfig):
        assert self.band == config.band
        assert self.sampling_rate == config.sampling_rate
        assert self.full_scale_power_dbm == config.power
        if config.upconverter not in self.upconverters:
            self.upconverters[config.upconverter] = {"frequency": config.frequency}
        else:
            assert (
                config.frequency == self.upconverters[config.upconverter]["frequency"]
            )


@dataclass(frozen=True)
class MwFemInput:
    downconverter_frequency: float
    band: int
    sampling_rate: float

    @classmethod
    def from_config(cls, config: MwFemOscillatorConfig):
        return cls(
            downconverter_frequency=config.frequency,
            band=config.band,
            sampling_rate=config.sampling_rate,
        )


@dataclass(frozen=True)
class OctaveOutput:
    LO_frequency: int
    gain: int = 0
    LO_source: Literal["internal", "external"] = "internal"
    output_mode: OctaveOutputModes = "triggered"

    @classmethod
    def from_config(cls, config: Union[OscillatorConfig, OctaveOscillatorConfig]):
        kwargs = dict(LO_frequency=config.frequency, gain=config.power)
        if isinstance(config, OctaveOscillatorConfig):
            kwargs["output_mode"] = config.output_mode
        return cls(**kwargs)


@dataclass(frozen=True)
class OctaveInput:
    LO_frequency: int
    LO_source: Literal["internal", "external"] = "internal"
    IF_mode_I: Literal["direct", "envelop", "mixer"] = "direct"
    IF_mode_Q: Literal["direct", "envelop", "mixer"] = "direct"


ModuleTypes = Literal["opx1", "LF", "MW"]


@dataclass
class Controller:
    type: ModuleTypes = "opx1"
    """https://docs.quantum-machines.co/latest/docs/Introduction/config/?h=opx10#controllers"""
    analog_outputs: dict[int, dict] = field(default_factory=dict)
    digital_outputs: dict[int, dict] = field(default_factory=dict)
    analog_inputs: dict[int, Union[AnalogInput, MwFemInput]] = field(
        default_factory=dict
    )

    def _set_default_inputs(self):
        """Add default inputs in controller config section.

        Inputs are always registered to avoid issues with automatic mixer
        calibration when using Octaves.
        """
        for port in range(1, 3):
            if port not in self.analog_inputs:
                self.analog_inputs[port] = {"offset": 0}

    def add_octave_output(self, port: int):
        # TODO: Add offset here?
        self._set_default_inputs()
        self.analog_outputs[2 * port - 1] = {"offset": 0}
        self.analog_outputs[2 * port] = {"offset": 0}

        self.digital_outputs[2 * port - 1] = {}

    def add_octave_input(self, port: int, config: QmAcquisitionConfig):
        self._set_default_inputs()
        self.analog_inputs[2 * port - 1] = self.analog_inputs[2 * port] = (
            AnalogInput.from_config(config)
        )


@dataclass
class Opx1000:
    type: Literal["opx1000"] = "opx1000"
    fems: dict[int, Controller] = field(default_factory=dict)


@dataclass
class Octave:
    connectivity: Union[str, tuple[str, int]]
    RF_outputs: dict[int, OctaveOutput] = field(default_factory=dict)
    RF_inputs: dict[int, OctaveInput] = field(default_factory=dict)

    def __post_init__(self):
        if "/" in self.connectivity:
            con, fem = self.connectivity.split("/")
            self.connectivity = (con, int(fem))


ControllerId = Union[str, tuple[str, int]]


def process_controller_id(id: ControllerId):
    """Convert controller identifier depending on cluster type.

    For OPX+ clusters ``id`` is just the controller name (eg. 'con1').
    For OPX1000 clusters ``id`` has the format
    '{controller_name}/{fem_number}' (eg. 'con1/4').
    In that case ``id`` may also be a ``tuple``
    `(controller_name, fem_number)`
    """
    if isinstance(id, tuple):
        con, fem = id
        return con, fem
    if "/" in id:
        con, fem = id.split("/")
        return con, int(fem)
    return id, None


class Controllers(dict[str, Union[Controller, Opx1000]]):
    """Dictionary of controllers compatible with OPX+ and OPX1000."""

    def __contains__(self, key: ControllerId) -> bool:
        con, fem = process_controller_id(key)
        contains = super().__contains__(con)
        if fem is None:
            return contains
        return contains and fem in self[con].fems

    def __getitem__(self, key: ControllerId) -> Controller:
        con, fem = process_controller_id(key)
        value = super().__getitem__(con)
        if fem is None:
            return value
        return value.fems[fem]

    def __setitem__(self, key: ControllerId, value: Controller):
        con, fem = process_controller_id(key)
        if fem is None:
            super().__setitem__(key, value)
        else:
            if con not in self:
                super().__setitem__(con, Opx1000())
            self[con].fems[fem] = value



import time
import warnings
from collections import defaultdict
from functools import cached_property
from itertools import groupby
from typing import Optional, cast

import qblox_instruments as qblox
from qblox_instruments.qcodes_drivers.module import Module
from qcodes.instrument import find_or_create_instrument

from qibolab._core.components import AcquisitionChannel, Configs, DcConfig, IqChannel
from qibolab._core.execution_parameters import AcquisitionType, ExecutionParameters
from qibolab._core.identifier import ChannelId, Result
from qibolab._core.instruments.abstract import Controller
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers, normalize_sweepers

from . import config
from .config import PortAddress
from .identifiers import SequencerMap, SlotId
from .log import Logger
from .results import AcquiredData, extract, integration_lenghts
from .sequence import Q1Sequence, compile
from .utils import batch_shots, concat_shots, lo_configs, time_of_flights
from .validate import assert_channels_exclusion, validate_sequence

__all__ = ["Cluster"]

SAMPLING_RATE = 1


class Cluster(Controller):
    """Controller object for Qblox cluster."""

    name: str
    """Device name.

    As described in:
    https://docs.qblox.com/en/main/getting_started/setup.html#connecting-to-multiple-instruments
    """
    bounds: str = "qblox/bounds"
    _cluster: Optional[qblox.Cluster] = None

    @property
    def cluster(self) -> qblox.Cluster:
        """Ensure cluster object access.

        The object presence results from an existing connection.
        """
        assert self._cluster is not None
        return self._cluster

    @property
    def sampling_rate(self) -> int:
        """Provide instrument's sampling rate."""
        return SAMPLING_RATE

    @property
    def is_connected(self) -> bool:
        """Determine connections status."""
        return self._cluster is not None

    def connect(self):
        """Connect and initialize the instrument."""
        if self.is_connected:
            return

        self._cluster = find_or_create_instrument(
            qblox.Cluster, recreate=True, name=self.name, identifier=self.address
        )
        self._cluster.reset()

    def disconnect(self):
        """Disconnect and reset the instrument."""
        assert self._cluster is not None

        for module in self._modules.values():
            module.stop_sequencer()
        self._cluster.reset()
        self._cluster.close()
        self._cluster = None

    def play(
        self,
        configs: Configs,
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ) -> dict[int, Result]:
        """Execute the given experiment."""
        results = {}
        log = Logger(configs)

        # no unrolling yet: act one sequence at a time, and merge results
        for ps in sequences:
            # split shots in batches, in case the required experiment exceeds the
            # allowed memory
            psres = []
            for shots in batch_shots(ps, sweepers, options):
                options_ = options.model_copy(update={"nshots": shots})
                sweepers_ = normalize_sweepers(
                    sweepers,
                    lo_configs(self._los, configs),
                    {
                        ch: cfg.offset
                        for ch, cfg in configs.items()
                        if isinstance(cfg, DcConfig)
                    },
                )
                # first compile pulses and sweepers into Qblox sequences
                assert_channels_exclusion(ps, self._probes)
                sequences_ = compile(
                    ps,
                    sweepers_,
                    options_,
                    self.sampling_rate,
                    time_of_flights(configs),
                )
                for seq in sequences_.values():
                    validate_sequence(seq)
                log.sequences(sequences_)

                # then configure modules and sequencers
                # (including sequences upload)
                sequencers = self._configure(
                    sequences_, configs, options_.acquisition_type
                )
                log.status(self.cluster, sequencers)

                # finally execute the experiment, and fetch results
                duration = options_.estimate_duration([ps], sweepers_)
                data = self._execute(
                    sequencers, sequences_, duration, options_.acquisition_type
                )
                log.data(data)

                # process raw results to adhere to standard format
                lenghts = integration_lenghts(sequences_, sequencers, self._modules)
                psres.append(
                    extract(
                        data,
                        lenghts,
                        options_.acquisition_type,
                        options_.results_shape(sweepers_),
                    )
                )

            # update results - concatenating shots, if needed
            results |= concat_shots(psres, options)
        return results

    def _configure(
        self,
        sequences: dict[ChannelId, Q1Sequence],
        configs: Configs,
        acquisition: AcquisitionType,
    ) -> SequencerMap:
        """Configure modules and sequencers.

        The return value consists of the association map from channels
        to sequencers, for each module.
        """
        sequencers = defaultdict(dict)
        for slot, chs in self._channels_by_module.items():
            module = self._modules[slot]

            # each channel is going to be assigned to its own sequencer, thus they can
            # not be outnumbered
            assert len(module.sequencers) >= len(chs)

            # compute module configurations, and apply them
            los = config.module.los(self._los, configs, chs)
            config.ModuleConfig.compute(self.channels, los, module.is_qrm_type).apply(
                module
            )

            # configure all sequencers, and store active ones' association to channels
            rf = module.is_rf_type
            for idx, ((ch, address), sequencer) in enumerate(
                zip(chs, module.sequencers)
            ):
                seq = sequences.get(ch, Q1Sequence.empty())
                config.SequencerConfig.build(
                    address, seq, ch, self.channels, configs, acquisition, idx, rf
                ).apply(sequencer)
                # only collect active sequencers
                if not seq.is_empty:
                    sequencers[slot][ch] = idx

        return sequencers

    def _execute(
        self,
        sequencers: SequencerMap,
        sequences: dict[ChannelId, Q1Sequence],
        duration: float,
        acquisition: AcquisitionType,
    ) -> dict[ChannelId, AcquiredData]:
        """Execute experiment and fetch results."""
        # start experiment, arming and starting all sequencers
        for mod, seqs in sequencers.items():
            module = self._modules[mod]
            for seq in seqs.values():
                module.arm_sequencer(seq)
            module.start_sequencer()

        # approximately wait for experiment completion
        time.sleep(duration + 1)

        # fetch acquired results
        acquisitions = {}
        for slot, seqs in sequencers.items():
            for ch, seq in seqs.items():
                # wait all sequencers
                status = self.cluster.get_sequencer_status(slot, seq, timeout=1)
                if status.status is qblox.SequencerStatuses.ERROR:
                    raise RuntimeError(f"slot: {slot}, seq: {seq}\n{status}")
                if status.status is qblox.SequencerStatuses.WARNING:
                    warnings.warn(f"slot: {slot}, seq: {seq}\n{status}")

                # skip results retrieval for passive or inactive sequencers...
                sequence = sequences.get(ch)
                if sequence is None:
                    continue
                # ... and also for channels missing acquisition instructions
                seq_acqs = sequence.acquisitions
                if len(seq_acqs) == 0:
                    continue

                # ensure acquisition status, and fetch results
                self.cluster.get_acquisition_status(slot, seq, timeout=1)
                if acquisition is AcquisitionType.RAW:
                    for name in seq_acqs:
                        self.cluster.store_scope_acquisition(slot, seq, str(name))
                acquisitions[ch] = self.cluster.get_acquisitions(slot, seq)

        return acquisitions

    @cached_property
    def _modules(self) -> dict[SlotId, Module]:
        """Retrieve slot to module object mapping."""
        return {mod.slot_idx: mod for mod in self.cluster.modules if mod.present()}

    @cached_property
    def _probes(self) -> set[ChannelId]:
        """Determine probe channels."""
        return {
            ch.probe
            for ch in self.channels.values()
            if isinstance(ch, AcquisitionChannel) and ch.probe is not None
        }

    @cached_property
    def _channels_by_module(self) -> dict[SlotId, list[tuple[ChannelId, PortAddress]]]:
        """Identify channels associated to each module.

        Channels are otherwise a set, where the association is stored in
        the :attr:`qibolab.Channel.port` attributes of each channel.
        """
        addresses = {
            name: PortAddress.from_path(ch.path) for name, ch in self.channels.items()
        }
        return {
            k: [el[1] for el in g]
            for k, g in groupby(
                sorted(
                    (address.slot, (ch, address))
                    for ch, address in addresses.items()
                    if ch not in self._probes
                ),
                key=lambda el: el[0],
            )
        }

    @cached_property
    def _los(self) -> dict[ChannelId, str]:
        """Extract channel to LO mapping.

        The result contains the associated channel, since required to
        address the LO through the API. While the LO identifier is used
        to retrieve the configuration.
        """
        channels = self.channels
        return {
            ch: lo
            for ch, lo in (
                (ch, cast(IqChannel, channels[iq]).lo)
                for ch, iq in ((ch, channels[ch].iqout(ch)) for ch in self.channels)
                if iq is not None
            )
            if lo is not None
        }



"""Pulse class."""

from typing import Annotated, Literal, Union, cast
from uuid import uuid4

import numpy as np
from pydantic import UUID4, Field

from ..serialize import Model
from .envelope import Envelope, IqWaveform, Waveform

__all__ = [
    "Acquisition",
    "Align",
    "Delay",
    "Pulse",
    "PulseId",
    "PulseLike",
    "Readout",
    "VirtualZ",
]

PulseId = UUID4
"""Unique identifier for a pulse."""


class _PulseLike(Model):
    id_: PulseId = Field(default_factory=uuid4, exclude=True)

    @property
    def id(self) -> PulseId:
        """Instruction identifier."""
        return self.id_

    def new(self) -> "PulseLike":
        return cast(PulseLike, self.model_copy(deep=True, update={"id_": uuid4()}))

    def __eq__(self, other: object) -> bool:
        """Compare instances."""
        # TODO: for the time being, Pydantic does not support fields
        # inclusion/exclusion in comparison, otherwise it would be much better to
        # exclude them rather then applying this recursive definition
        # https://github.com/pydantic/pydantic/discussions/6717
        s = vars(self)
        o = vars(other)
        return isinstance(other, type(self)) and all(
            s[k] == o[k] for k in s if k != "id_"
        )

    def __hash__(self) -> int:
        return hash(tuple(v for k, v in vars(self).items() if k != "id_"))


class Pulse(_PulseLike):
    """A pulse to be sent to the QPU.

    Valid on any channel, except acquisition ones.
    """

    kind: Literal["pulse"] = "pulse"

    duration: float
    """Pulse duration."""

    amplitude: float
    """Pulse digital amplitude (unitless).

    Pulse amplitudes are normalised between -1 and 1.
    """
    envelope: Envelope
    """The pulse envelope shape.

    See :class:`qibolab.Envelope` for list of available shapes.
    """
    relative_phase: float = 0.0
    """Relative phase of the pulse, in radians."""

    def i(self, sampling_rate: float) -> Waveform:
        """Compute the envelope of the waveform i component."""
        samples = int(self.duration * sampling_rate)
        return self.amplitude * self.envelope.i(samples)

    def q(self, sampling_rate: float) -> Waveform:
        """Compute the envelope of the waveform q component."""
        samples = int(self.duration * sampling_rate)
        return self.amplitude * self.envelope.q(samples)

    def envelopes(self, sampling_rate: float) -> IqWaveform:
        """Compute a tuple with the i and q envelopes."""
        return np.array([self.i(sampling_rate), self.q(sampling_rate)])


class Delay(_PulseLike):
    """Wait instruction.

    During its length no pulse is sent on the same channel.

    Valid on any channel.
    """

    kind: Literal["delay"] = "delay"

    duration: float
    """Duration in ns."""


class VirtualZ(_PulseLike):
    """Implementation of Z-rotations using virtual phase.

    Only valid on a drive channel.
    """

    kind: Literal["virtualz"] = "virtualz"

    phase: float
    """Phase that implements the rotation."""

    @property
    def duration(self):
        """Duration of the virtual gate should always be zero."""
        return 0


class Acquisition(_PulseLike):
    """Acquisition instruction.

    This event instructs the device to acquire samples for the event
    span.

    Only valid on an acquisition channel.
    """

    kind: Literal["acquisition"] = "acquisition"

    duration: float
    """Duration in ns."""


class Readout(_PulseLike):
    """Readout instruction.

    This event instructs the device to acquire samples for the event
    span.

    Only valid on an acquisition channel.
    """

    kind: Literal["readout"] = "readout"

    acquisition: Acquisition
    probe: Pulse

    @classmethod
    def from_probe(cls, probe: Pulse):
        """Create a whole readout operation from its probe pulse.

        The acquisition is made to match the same probe duration.
        """
        return cls(acquisition=Acquisition(duration=probe.duration), probe=probe)

    @property
    def duration(self) -> float:
        """Duration in ns."""
        return self.acquisition.duration

    @property
    def id(self) -> PulseId:
        """Instruction identifier."""
        return self.acquisition.id

    def new(self) -> "PulseLike":
        return cast(
            PulseLike,
            self.model_copy(
                deep=True,
                update={
                    "id_": uuid4(),
                    "acquisition": self.acquisition.new(),
                    "probe": self.probe.new(),
                },
            ),
        )


class Align(_PulseLike):
    """Brings different channels at the same point in time."""

    kind: Literal["align"] = "align"


PulseLike = Annotated[
    Union[Align, Pulse, Delay, VirtualZ, Acquisition, Readout],
    Field(discriminator="kind"),
]



import re
from collections import defaultdict
from typing import Optional, Union

from qibolab._core.components.channels import (
    AcquisitionChannel,
    Channel,
    DcChannel,
    IqChannel,
)
from qibolab._core.identifier import QubitId

__all__ = ["infer_los", "map_ports"]


def _chtype(mod: str, input: bool) -> tuple[str, type[Channel]]:
    if mod.startswith("qcm_rf"):
        return "drive", IqChannel
    if mod.startswith("qcm"):
        return "flux", DcChannel
    if mod.startswith("qrm_rf"):
        if input:
            return "acquisition", AcquisitionChannel
        return "probe", IqChannel
    raise ValueError


def _port_channels(mod: str, port: Union[int, str], slot: int) -> dict:
    if isinstance(port, str) and port.startswith("io"):
        return {
            "probe": IqChannel(path=f"{slot}/o{port[2:]}"),
            "acquisition": AcquisitionChannel(path=f"{slot}/i{port[2:]}"),
        }
    port = f"o{port}" if isinstance(port, int) else port
    name, cls = _chtype(mod, port[0] == "i")
    return {name: cls(path=f"{slot}/{port}")}


def _premap(cluster: dict):
    d = defaultdict(lambda: defaultdict(dict))

    for mod, props in cluster.items():
        slot = props[0]
        for port, els in props[1].items():
            for el in els:
                d[el] |= _port_channels(mod, port, slot)

    return d


def map_ports(cluster: dict, qubits: dict, couplers: Optional[dict] = None) -> dict:
    """Extract channels from compact representation.

    Conventions:
    - each item is a module
    - the first element of each value is the module's slot ID
    - the second element is a map from ports to qubits
        - ports
            - they are `i{n}` or `o{n}` for the inputs and outputs respectively
            - `io{n}` is also allowed, to signal that both are connected (cater for the specific
              case of the QRM_RF where there are only one port of each type)
            - if it's just an integer, it is intended to be an output (equivalent to `o{n}`)
        - values
            - list of element names
            - they are `q{name}` or `c{name}` for qubits and couplers respectively
            - multiple elements are allowed, for multiplexed ports

    .. note::

        integer qubit names are not allowed

    .. todo::

        Currently channel types are inferred from the module type, encoded in its name. At
        least an override should be allowed (per module, per port, per qubit).
    """
    if couplers is None:
        couplers = {}

    premap = _premap(cluster)

    channels = {}
    for name, el in (qubits | couplers).items():
        for chname, ch in premap[name].items():
            channels[getattr(el, chname)] = ch
        try:
            channels[el.acquisition] = channels[el.acquisition].model_copy(
                update={"probe": el.probe}
            )
        except KeyError:
            pass
    return channels


_PORT = re.compile(r"i?o?(.*)")


def _qubit_id(string: str) -> QubitId:
    try:
        return int(string)
    except ValueError:
        return string


def _digits(string: str) -> QubitId:
    res = re.search(_PORT, string)
    return _qubit_id(res[1]) if res is not None else ""


def _out_port(port: Union[str, int]) -> QubitId:
    return port if isinstance(port, int) else _digits(port)


def infer_los(cluster: dict) -> dict[tuple[QubitId, bool], str]:
    """Infer LOs names for output channels.

    ``cluster`` should be a mapping compatible with the same input of :func:`map_ports`.

    The result is a mapping from ``(qubit, channel)``, where ``qubit`` is the identifier,
    and ``channel`` is a boolean toggle: ``True`` for probe channels, ``False`` for
    drive.
    """
    return {
        (q, "qrm" in mod): f"{mod}/o{_out_port(port)}/lo"
        for mod, specs in cluster.items()
        if "_rf" in mod
        for port, qs in specs[1].items()
        for q in qs
    }
